"""
A module that assists with implementing the Discord OAuth2 flow.
"""
import collections
import datetime
import enum
import secrets
import typing

from asks.sessions import Session
from oauthlib.oauth2 import OAuth2Error
from oauthlib.oauth2.rfc6749.clients import WebApplicationClient


class InvalidStateError(Exception):
    """
    Raised from :meth:`.OAuth2Handshaker.fetch_token` if the state is invalid
    """
    pass


class OAuth2Scope(enum.Enum):
    """
    OAuth2 scopes.
    """
    #: Authorizes a bot into a guild.
    BOT = 'bot'
    #: Allows access to the connections of a user.
    CONNECTIONS = 'connections'
    #: Allows access to basic user info.
    IDENTIFY = 'identify'
    #: Allows access to basic user info + their email.
    EMAIL = 'email'
    #: Allows access to user guild objects for this user.
    GUILDS = 'guilds'
    #: Allows forcibly joining a guild on the behalf of a user.
    GUILDS_JOIN = 'guilds.join'
    #: Allows forcibly adding users to group DMs.
    GDM_JOIN = 'gdm.join'

    #: Allows reading messages via RPC.
    MESSAGES_READ = 'messages.read'
    #: Allows RPC client control.
    RPC = 'rpc'
    #: Allows RPC API control.
    RPC_API = 'rpc.api'
    #: Allows RPC notification reading.
    RPC_NOTIFICATIONS_READ = 'rpc.notifications.read'

    #: Creates an incoming webhook when authorized.
    WEBHOOK_INCOMING = 'webhook.incoming'


class OAuth2Token(object):
    """
    Represents a token returned from the Discord OAuth2 API.
    """

    def __init__(self, token_type: str, scope: str,
                 access_token: str, refresh_token: str, expiration_time: datetime.datetime):
        #: The token type of the token (normally ``Bearer``).
        self.token_type = token_type

        #: A list of :class:`~.OAuth2Scope` this token is authenticated for.
        self.scopes = []
        for scope_name in scope.split(" "):
            self.scopes.append(OAuth2Scope(scope_name))

        #: The actual access token to be used.
        self.access_token = access_token

        #: The refresh token to be used during a refresh.
        self.refresh_token = refresh_token

        #: The time this token expires at.
        self.expiration_time = expiration_time

    @classmethod
    def from_dict(cls, d: dict) -> 'OAuth2Token':
        """
        Creates a token from a dict, similar to one provided by the token endpoint.
        """
        if "expiration_time" not in d:
            expiration_time = datetime.datetime.utcnow() + \
                              datetime.timedelta(seconds=d["expires_in"])
            d["expiration_time"] = expiration_time

        d.pop("expires_in", None)

        c = cls(**d)
        return c

    @property
    def expired(self) -> bool:
        return self.expiration_time < datetime.datetime.utcnow()

    def __repr__(self) -> str:
        return "<OAuth2Token access='{}' refresh='{}'>".format(self.access_token,
                                                               self.refresh_token)


class OAuth2Client(object):
    """
    The class used to perform an OAuth2 handshake with Discord.
    This will provide a URL that can be used to authorize a client, and then the ability to fetch a
    new :class:`~.OAuth2Token`.
    
    .. code-block:: python3
    
        # note: we can't fetch the client ID automatically, unlike in regular bots :(
        # so you have to pass it manually
        # if you have a bot running alongside this, you can use the Client ID from that, however.
        my_client = OAuth2Client(my_client_id, my_client_secret)
        
        url = my_client.get_authorization_url(scopes=[OAuthScope.IDENTITY])
        # get the user to give back the state and the code
        # ...
        token = await my_client.get_token(state, code)
        
        # some time later
        user = await my_client.get_user()

    
    :param client_id:
        The Client ID of the application.
        
        .. warning::
            For **old bots**, this is NOT the Bot ID.
     
    :param client_secret: 
        The client secret of the application.
        
        ..warning::
            This is **not** your token.
            
    :param redirect_uri: 
        The URL the client will be redirected to after authorizing your application.  
        This is then used to retrieve the state.
    """

    BASE = "https://discordapp.com"
    API_BASE = "/api/v7"
    AUTHORIZE_URL = "/oauth2/authorize"
    TOKEN_URL = "/oauth2/token"

    def __init__(self, client_id: int, client_secret: str,
                 redirect_uri: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

        self._oauth2_client = WebApplicationClient(client_id=self.client_id,
                                                   redirect_url=self.redirect_uri)
        self.sess = Session(base_location=self.BASE, endpoint=self.API_BASE)

        #: A list of states that have been seen before.
        #: If the state was not seen, it will raise an invalid state error.
        self._states = collections.deque(maxlen=500)

    def _get_state(self):
        return secrets.token_urlsafe(16)

    def get_authorization_url(self, scopes: typing.List[OAuth2Scope]):
        """
        Gets the authorization URL to be used for the user to authorize this application.
        
        :param scopes: A list of :class:`~.OAuthScope` for the request. 
        """
        url = self._oauth2_client.prepare_request_uri(self.AUTHORIZE_URL,
                                                      scope=[scope.value for scope in scopes],
                                                      redirect_uri=self.redirect_uri,
                                                      state=self._get_state())
        return url

    async def fetch_token(self, code: str, state: str) -> OAuth2Token:
        """
        Fetches the token when given an authorization code and state.
        
        :param code: The authorization code returned in the URI. 
        :param state: The state returned in the URI.
        :return: A :class:`~.OAuthToken` object representing the token.
        """
        # if state not in self._states:
        #    raise InvalidStateError(state)
        # construct the URI we need to send
        uri = self._oauth2_client.prepare_request_uri(self.TOKEN_URL,
                                                      code=code,
                                                      redirect_uri=self.redirect_uri,
                                                      client_secret=self.client_secret,
                                                      grant_type="authorization_code")
        response = await self.sess.post(path=uri)
        if response.status_code != 200:
            raise OAuth2Error(response, response.json())

        # construct the token
        token = OAuth2Token.from_dict(response.json())
        return token

    async def refresh_token(self, token: typing.Union[OAuth2Token, str],
                            scopes: typing.List[OAuth2Scope] = None) -> OAuth2Token:
        """
        Refreshes a token.
        
        :param token: Either a :class:`~.OAuth2Token` or the str refresh token.
        :param scopes: The scopes to request.
            If an OAuth2Token is passed as the token, this will be fetched automatically.
        :return: The refreshed :class:`~.OAuth2Token`.
        """
        if isinstance(token, OAuth2Token):
            ref = token.refresh_token
            scopes = [scope.name.lower() for scope in token.scopes]
        else:
            ref = token
            scopes = [scope.name.lower() for scope in scopes] if scopes else None

        uri, headers, body = self._oauth2_client.prepare_refresh_token_request(
            self.TOKEN_URL, refresh_token=ref, scope=scopes
        )

        response = await self.sess.post(path=uri, headers=list(headers.items()), body=body)

        self._oauth2_client.parse_request_body_response(body=await response.text())
        token = OAuth2Token.from_dict(await response.json())
        return token



"""
A voice player that uses ffmpeg.
"""
import subprocess
import threading
import typing

import time

from curious.voice import voice_client


class VoicePlayer(threading.Thread):
    """
    A voice player object.
    """

    FFMPEG_NAME = "ffmpeg"

    def __init__(self, vc: 'voice_client.VoiceClient',
                 path: str, callback: typing.Callable[['VoicePlayer'], None] = None):
        """
        :param vc: The voice client this player is associated with.
        :param path: The path this player should play.
        :param callback: A callback to be ran after the voice player has finished playing.
        """
        threading.Thread.__init__(self, daemon=True)
        self.vc = vc
        self.path = path
        self.callable = callback

        # samples_per_frame * 4
        self.buf_amount = 20 * 48 * 4

        # delay - 20ms
        self.delay = 0.02

    def run(self):
        # Load ffmpeg, and begin encoding the frames.
        arg = "-i {} -f s16le -ar 48000 -ac 2 pipe:1".format(self.path).split()
        full = [self.FFMPEG_NAME, *arg]

        proc = subprocess.Popen(full, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        start_time = time.time()
        loops = 0

        while True:
            # Ensure the subprocess isn't ded
            try:
                proc.wait(timeout=0)
            except subprocess.TimeoutExpired:
                pass
            else:
                _, stderr = proc.communicate()
                if proc.returncode:
                    raise RuntimeError(stderr.decode())
                else:
                    break

            loops += 1

            # encode the packet and send it to discord
            data = proc.stdout.read(self.buf_amount)
            if not data:
                continue

            self.vc.send_voice_packet(data)

            # now we need to sleep
            # make sure to account for drift!
            delay = max(0, self.delay + ((start_time + self.delay * loops) + - time.time()))
            time.sleep(delay)

        if self.callable:
            self.callable(self)



"""
Base classes that all dataclasses inherit from.

.. currentmodule:: curious.dataclasses.bases
"""
import datetime

from curious.core import client

DISCORD_EPOCH = 1420070400000


class IDObject(object):
    """
    This object is comparable using the snowflake as an ID.

    It is also hashable, using the ID as a hash.
    """

    __slots__ = "id",

    def __init__(self, id: int):
        """
        :param id: The snowflake ID of the object.
        """
        if isinstance(id, str):
            id = int(id)

        #: The ID of this object.
        self.id = id

    def __repr__(self):
        return "<{} id={}>".format(self.__class__.__name__, self.id)

    __str__ = __repr__

    @property
    def snowflake_timestamp(self) -> datetime.datetime:
        """
        :return: The timestamp of the snowflake.
        """
        return datetime.datetime.utcfromtimestamp(((int(self.id) >> 22) + DISCORD_EPOCH) / 1000)

    def __eq__(self, other):
        if not hasattr(other, "id"):
            return NotImplemented

        return other.id == self.id

    def __hash__(self):
        return hash(self.id)


class Dataclass(IDObject):
    """
    The base class for all dataclasses.

    These contain a reference to the current bot as `_bot`.
    """

    # __weakref__ is used to allow weakreffing
    __slots__ = "_bot", "__weakref__"

    def __init__(self, id: int, cl: 'client.Client'):
        super().__init__(id)

        self._bot = cl



"""
Wrappers for Application Info objects.

.. currentmodule:: curious.dataclasses.appinfo
"""
from collections import namedtuple

from curious.dataclasses import guild as dt_guild
from curious.dataclasses.bases import Dataclass
from curious.exc import CuriousError

AuthorizedApp = namedtuple("AuthorizedApp", "scopes id application")


class AppInfo(Dataclass):
    """
    Represents the application info for an OAuth2 application.
    """

    def __init__(self, client, **kwargs):
        self._application = kwargs.get("application", {})

        #: The client ID of this application.
        self.client_id = int(self._application.get("id", 0))
        super().__init__(self.client_id, client)

        if "owner" in self._application:
            owner = self._bot.state.make_user(self._application.get("owner"))
        else:
            owner = None

        #: The owner of this application.
        #: This can be None if the application fetched isn't the bot's.
        self.owner = owner

        #: The name of this application.
        self.name = self._application.get("name", None)  # type: str

        #: The description of this application.
        self.description = self._application.get("description", None)  # type: str

        #: Is this bot public?
        self.public = self._application.get("bot_public", None)  # type: bool

        #: Does this bot require OAuth2 Code Grant?
        self.requires_code_grant = self._application.get("bot_require_code_grant",
                                                         None)  # type: bool

        #: The icon hash for this application.
        self._icon_hash = self._application.get("icon", None)  # type: str

        if "bot" in kwargs:
            #: The bot :class:`~.User` associated with this application.
            self.bot = self._bot.state.make_user(kwargs.get("bot", {}))  # type: User
        else:
            self.bot = None

    def __repr__(self):
        return "<{} owner='{}' name='{}' bot='{}'>".format(type(self).__name__, self.owner,
                                                           self.name, self.bot)

    @property
    def icon_url(self):
        """
        :return: The icon url for this bot.
        """
        if self._icon_hash is None:
            return None

        return "https://cdn.discordapp.com/app-icons/{}/{}.jpg".format(self.client_id,
                                                                       self._icon_hash)

    async def add_to_guild(self, guild: 'dt_guild.Guild', *,
                           permissions: int=0):
        """
        Authorizes this bot to join a guild.

        This requires a userbot client to be used.
        """
        if self._bot.is_bot:
            raise CuriousError("Bots cannot add other bots")

        if self.bot is None:
            raise CuriousError("This application has no bot associated")

        if self.owner is None and not self.public:
            raise CuriousError("This bot is not public")

        if self.requires_code_grant:
            raise CuriousError("This bot requires code grant")

        await self._bot.http.authorize_bot(self.client_id, guild.id, permissions=permissions)


class OwnedAppInfo(AppInfo):
    """
    A special class for application info
    """



"""
Represents a voice client.

Other documentation:

- Timestamp increases by 48 * 20 every time (samples * frame length).
- Sequence and timestamp both overflow.

frame_length = 20 (??)
sample_size = 4 (2 * 2)
samples_per_frame: 48 * 20
frame_size = samples_per_frame APPARENTLY


"""
import logging
import socket
import struct
from typing import Tuple

import curio
from curio.socket import getaddrinfo

try:
    from nacl.secret import SecretBox
except ImportError:
    has_nacl = False

try:
    from opuslib import Decoder, Encoder
except Exception:
    has_opus = False

from curious.voice import voice_player as vp
from curious.voice.voice_gateway import VoiceGateway

logger = logging.getLogger("curious.voice")


def simulate_overflow(value: int, bits: int, signed: bool):
    """
    Simulates integer overflow.
    """
    base = 1 << bits
    value %= base
    return value - base if signed and value.bit_length() == bits else value


class VoiceClient(object):
    """
    The voice client instance controls connecting to Discord's voice servers.

    This should ***not*** be created directly - instead use :class:`~.Channel.connect()` to connect
    to a voice channel, and use the instance returned from.
    """

    def __new__(cls, *args, **kwargs):
        if not has_opus or not has_nacl:
            raise RuntimeError("Cannot make new VoiceClients without libopus installed")

        return super().__new__(*args, **kwargs)

    def __init__(self, main_client,
                 channel):
        """
        :param main_client: The :class:`~.curious.core.client.Client` object associated with this
            VoiceClient.
        :param channel: The :class:`~.Channel` associated with this VoiceClient.
        """

        #: The main client this voice client is associated with.
        self.client = main_client

        #: The voice channel this client is connected to.
        self.channel = channel

        #: The voice websocket that we are connected to.
        self.vs_ws = None  # type: VoiceGateway

        #: The UDP socket that we are connected to.
        self._sock = None  # type: socket.socket

        self.main_task = None  # type: curio.Task

        # Header related stuff
        self.sequence = 0  # sequence is 2 bytes
        self.timestamp = 0  # timestamp is 4 bytes

        # Opus encoder/decoder
        # Do not use
        self.encoder = Encoder(48000, 2, 'audio')
        self.encoder.bitrate = 96000
        self.decoder = Decoder(48000, 2)

    @property
    def open(self):
        return self.vs_ws._open

    # Voice encoder related things.
    def get_packet_header(self) -> bytes:
        """
        Gets the voice packet header.

        :return: The bytes of the header.
        """
        header = bytearray(12)

        # constant values, provided by the docs
        header[0:2] = b"\x80\x78"

        # dynamic values
        # offset 2 -> sequence
        struct.pack_into(">H", header, 2, self.sequence)
        # offset 4 -> timestamp
        struct.pack_into(">I", header, 4, self.timestamp)
        # offset 8 -> ssrc
        struct.pack_into(">I", header, 8, self.vs_ws.ssrc)

        return header

    def get_voice_packet(self, opus_body: bytes) -> bytes:
        """
        Gets the voice packet to send to Discord, after encryption.

        :param opus_body: The body of the packet to encrypt.
        :return: The bytes of the packet.
        """
        header = self.get_packet_header()
        nonce = bytearray(24)
        # copy the header into nonce
        nonce[:12] = header

        encryptor = SecretBox(bytes(self.vs_ws.secret_key))
        encrypted_body = encryptor.encrypt(opus_body, nonce=bytes(nonce))

        # pack_nonce is True, so the body can just be concatted
        return bytes(header + encrypted_body.ciphertext)

    def get_ip_discovery_packet(self) -> bytes:
        """
        Gets the IP discovery packet to send to Discord.
        """
        packet = bytearray(70)
        packet[0:4] = struct.pack(">I", self.vs_ws.ssrc)

        return packet

    def unpack_packet(self, data: bytes) -> Tuple[int, int, int, bytes]:
        """
        Unpacks a voice packet received from Discord.

        :param data: The data to unpack.
        :return: A tuple of (ssrc, sequence, timestamp, data).
        """
        header = data[:12]
        encrypted_data = data[12:]

        # unpack header data
        type_ = header[0]
        version = header[1]
        sequence = struct.unpack(">H", header[2:4])[0]
        timestamp = struct.unpack(">I", header[4:8])[0]
        ssrc = struct.unpack(">I", header[8:12])[0]

        # okay, for some reason discord sends malformed packets
        # 0x90 as type means we need to chop off the first 8 bytes from the decrypted data
        # because it's invalid opus
        # first, decrypt the data

        nonce = bytearray(24)
        nonce[:12] = header

        encryptor = SecretBox(bytes(self.vs_ws.secret_key))
        decrypted = encryptor.decrypt(encrypted_data, nonce=bytes(nonce))

        if type_ == 0x90:
            decrypted = decrypted[8:]

        pcm_frames = self.decoder.decode(decrypted, 960)

        return ssrc, sequence, timestamp, pcm_frames

    def _send_voice_packet(self, built_packet: bytes):
        """
        Sends a voice packet.

        :param built_packet: The final built packet, as got by :meth:`.get_voice_packet`.
        """
        # Overflow values as appropriate.
        sequence = self.sequence + 1
        self.sequence = simulate_overflow(sequence, 2 * 8, False)

        timestamp = self.timestamp + (20 * 48)
        self.timestamp = simulate_overflow(timestamp, 2 * 16, False)

        try:
            self._sock.sendto(built_packet, (self.vs_ws.endpoint, self.vs_ws.port))
        except BlockingIOError:
            # can't send rn, oh well
            # opus is built for this
            logger.error("Failed to send voice packet! Sequence: {}, timestamp: {}".
                         format(self.sequence, self.timestamp))

    def send_opus_data(self, opus_data: bytes):
        """
        Sends an opus-encoded voice packet.

        :param opus_data: The data to send.
        """
        packet = self.get_voice_packet(opus_data)
        self._send_voice_packet(packet)

    def send_voice_packet(self, voice_data: bytes):
        """
        Sends voice data from a PCM buffer.

        :param voice_data: The raw PCM voice data to send.
        """
        data = self.encoder.encode(voice_data, 960)
        self.send_opus_data(data)

    def play_path(self, path: str):
        """
        Plays a file using ffmpeg.
        """
        player = vp.VoicePlayer(self, path)
        player.start()

    async def poll(self):
        """
        Polls the voice websocket constantly for new events.
        """
        while True:
            await self.vs_ws.next_event()

    async def close(self):
        """
        Closes the websocket. 
        """
        await self.vs_ws._close()
        await self.main_task.cancel()

    async def connect(self, timeout: int = 10) -> None:
        """
        Connects the voice client to the UDP server.

        :param timeout: The timeout before the connection is closed.
        """
        logger.info("Opening UDP connection to Discord voice servers.")
        # Wait before the voice socket is ready.
        await self.vs_ws._ready.wait()
        # Open our UDP connection.
        # TODO: Make this IPv6 compatible, if appropriate.
        # Now, you may be wondering why we use raw UDP sockets instead of using "nice" sockets.
        # This is because we need a voice thread to be able to access it.
        # This prevents blocking actions from killing the entire voice connection.

        addrinfo = await getaddrinfo(self.vs_ws.endpoint,
                                     self.vs_ws.port,
                                     0,
                                     socket.SOCK_DGRAM)

        for item in addrinfo:
            try:
                new_socket = socket.socket(item[0], item[1])
                break
            except:
                logger.debug("Failed to make socket", exc_info=True)
        else:
            raise ConnectionError("Could not create voice socket")

        self._sock = new_socket

        # Send an IP discovery packet.
        logger.info("Connecting to {}:{}".format(self.vs_ws.endpoint, self.vs_ws.port))
        packet = self.get_ip_discovery_packet()
        logger.debug("Sending IP discovery packet")
        self._sock.sendto(packet, (self.vs_ws.endpoint, self.vs_ws.port))

        # Wait for our local IP to be received.
        try:
            packet_data, addr = await curio.timeout_after(timeout,
                                                          curio.abide(self._sock.recvfrom, 70))
        except curio.TaskTimeout:
            self._sock.close()
            self.vs_ws._close()
            raise
        logger.debug("Got IP discovery packet!")

        # IP is encoded in ASCII, from the forth byte to the first \x00 byte.
        # Find the index of the null byte, starting from 4th.
        ip_start = 4
        ip_end = packet_data.index(0, ip_start)
        our_ip = packet_data[ip_start:ip_end].decode('ascii')

        # Also, we need our local port.
        # Unpack the little-endian (thanks discord!) port from the very end of the voice packet.
        our_port = struct.unpack_from('<H', packet_data, len(packet_data) - 2)[0]

        # Ask the voice websocket to send our SESSION DESCRIPTION packet.
        await self.vs_ws.send_select_protocol(our_ip, our_port)
        await self.vs_ws._got_secret_key.wait()
        logger.info("Established connection to Discord's voice servers.")

        self._sock.setblocking(False)

        # We are DONW!

    @classmethod
    async def create(cls, main_client,
                     gateway, channel) -> 'VoiceClient':
        """
        Creates a new VoiceClient from a channel and a gateway instance.

        :param main_client: The main :class:`.Client` to use.
        :param gateway: The gateway instance to use.
        :param channel: The :class:`~.Channel` to connect to.
        """
        vs_ws = await VoiceGateway.from_gateway(gw=gateway, channel_id=channel.id,
                                                guild_id=channel.guild.id)
        obb = cls(main_client, channel=channel)
        obb.vs_ws = vs_ws
        obb.main_task = await curio.spawn(obb.poll())
        return obb



"""
Curious - A Curio-based Python 3.5+ library for Discord bots.

.. currentmodule:: curious

.. autosummary::
    :toctree: 
    
    core
    commands
    dataclasses
    ext.loapi
    ext.paginator
    voice
    
    exc
    util
"""

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("discord-curious").version
except DistributionNotFound:
    __version__ = "0.0.0"


from curious.core.client import AUTOSHARD, BotType, Client
from curious.core.event import EventContext, event
from curious.core.gateway import Gateway
from curious.core.state import GuildStore, State
from curious.dataclasses.appinfo import AppInfo
from curious.dataclasses.bases import Dataclass, IDObject
from curious.dataclasses.channel import Channel, ChannelType
from curious.dataclasses.embed import Attachment, Embed
from curious.dataclasses.emoji import Emoji
from curious.dataclasses.guild import ContentFilterLevel, Guild, GuildChannelWrapper, \
    GuildEmojiWrapper, GuildRoleWrapper, MFALevel, NotificationLevel, VerificationLevel
from curious.dataclasses.invite import Invite, InviteChannel, InviteGuild, InviteMetadata
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.permissions import Overwrite, Permissions
from curious.dataclasses.presence import Game, Presence, Status
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.role import Role
from curious.dataclasses.search import SearchQuery, SearchResults
from curious.dataclasses.user import BotUser, User, UserProfile, UserSettings
from curious.dataclasses.voice_state import VoiceState
from curious.dataclasses.webhook import Webhook
from curious.dataclasses.widget import Widget, WidgetChannel, WidgetGuild, WidgetMember



"""
The core of Curious.

This package contains the bulk of the network interface with Discord, including parsing data that is incoming and 
delegating it to client code.

.. currentmodule:: curious.core

.. autosummary::
    :toctree: core
    
    client
    event
    gateway
    httpclient
    state
"""
import asks
import multio

if asks.init != multio.init:
    _init = multio.init
    multio.init = lambda lib: (asks.init(lib), _init(lib))

multio.init("curio")



"""
Misc utilities used in commands related things.
"""
import collections
import inspect
from typing import Callable, Iterable, List, Union

from curious.commands.exc import MissingArgumentError
from curious.core.client import Client
from curious.dataclasses.message import Message
from curious.util import replace_quotes


def get_full_name(func) -> str:
    """
    Gets the full name of a command, taking into account it's parents.
    """
    name = []
    # loop over and extract the name
    # then set func to the parent
    # and break if the parent is None
    while True:
        name.append(func.cmd_name)
        if func.cmd_parent is None:
            break

        func = func.cmd_parent

    return ' '.join(reversed(name))


async def _convert(ctx, tokens: List[str], signature: inspect.Signature):
    """
    Converts tokens passed from discord, using a signature.
    """
    final_args = []
    final_kwargs = {}

    args_it = iter(tokens)
    for n, (name, param) in enumerate(signature.parameters.items()):
        if n == 0:
            # Don't convert the `ctx` argument.
            continue

        assert isinstance(param, inspect.Parameter)
        # We loop over the signature parameters because it's easier to use those to consume.
        # Get the next argument from args.
        try:
            arg = next(args_it)
        except StopIteration as e:
            # not good!
            # If we're a *arg format, we can safely handle this, or if we have a default.
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                break

            if param.default is inspect.Parameter.empty:
                raise MissingArgumentError(ctx, param.name) from e

            # we continue immediately so that we dont try and do anything with defaults
            continue

        # Begin the consumption!
        if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          inspect.Parameter.POSITIONAL_ONLY]:
            # Only add it to the final_args, then continue the loop.
            arg = replace_quotes(arg)
            converter = ctx._lookup_converter(param.annotation)
            final_args.append(converter(ctx, arg))
            continue

        if param.kind in [inspect.Parameter.KEYWORD_ONLY]:
            # Only add it to final_kwargs.
            # This is a consume all operation, so we eat all of the arguments.
            f = [arg]

            while True:
                try:
                    next_arg = next(args_it)
                except StopIteration:
                    break

                f.append(next_arg)

            converter = ctx._lookup_converter(param.annotation)
            if len(f) == 1:
                final_kwargs[param.name] = converter(ctx, f[0])
            else:
                final_kwargs[param.name] = converter(ctx, " ".join(f))
            continue

        if param.kind in [inspect.Parameter.VAR_POSITIONAL]:
            # This *shouldn't* be called on `*` arguments, but we can't be sure.
            # Special case - consume ALL the arguments.
            f = [arg]

            while True:
                try:
                    next_arg = next(args_it)
                except StopIteration:
                    break

                f.append(next_arg)

            converter = ctx._lookup_converter(param.annotation)
            final_args.append(converter(ctx, " ".join(f)))

        if param.kind in [inspect.Parameter.VAR_KEYWORD]:
            # no
            continue

    return final_args, final_kwargs


def get_description(func) -> str:
    """
    Gets the description of a function.

    :param func: The function.
    :return: The description extracted from the docstring, or None.
    """
    if not func.__doc__:
        return None

    doc = inspect.cleandoc(func.__doc__)
    lines = doc.split("\n")
    return lines[0]


def get_usage(func, invoked_as: str = None) -> str:
    """
    :return: The usage text for this command.
    """
    if invoked_as is not None:
        final = [invoked_as]
    else:
        final = [func.cmd_name]

    signature = inspect.signature(func)

    for n, (name, param) in enumerate(signature.parameters.items()):
        # always skip the first arg, as it's self/ctx
        if n == 0:
            continue

        # check if we should skip the 2nd arg
        # not always possible
        from curious.commands.context import Context
        if name in ["ctx", "context"] or param.annotation is Context:
            continue

        # switch based on kind (again...)
        assert isinstance(param, inspect.Parameter)
        if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          inspect.Parameter.POSITIONAL_ONLY]:
            if param.annotation is not inspect.Parameter.empty:
                s = "<{}: {}>".format(param.name, param.annotation.__name__)
            else:
                s = "<{}>".format(param.name)

        elif param.kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_POSITIONAL]:
            if param.default is not inspect.Parameter.empty:
                if param.annotation is not inspect.Parameter.empty:
                    s = "[{}: {} (default: {})]".format(param.name, param.annotation.__name__,
                                                        repr(param.default))
                else:
                    s = "[{} (default: {})]".format(param.name, repr(param.default))
            else:
                if param.annotation is not inspect.Parameter.empty:
                    s = "<{}: {}>".format(param.name, param.annotation.__name__)
                else:
                    s = "<{}>".format(param.name)
        else:
            s = ""

        final.append(s)

    return " ".join(final)


def split_message_content(content: str, delim: str = " ") -> List[str]:
    """
    Splits a message into individual parts by `delim`, returning a list of strings.
    This method preserves quotes.

    .. code-block:: python3

        content = '!send "Fuyukai desu" "Hello, world!"'
        split = split_message_content(content, delim=" ")

    :param content: The message content to split.
    :param delim: The delimiter to split on.
    :return: A list of items split
    """

    # https://stackoverflow.com/a/43035638
    tokens = []
    cur = ''
    in_quotes = False

    for char in content.strip():
        if char == ' ' and not in_quotes:
            tokens.append(cur)
            cur = ''
        elif char == '"' and not in_quotes:
            in_quotes = True
            cur += char
        elif char == '"' and in_quotes:
            in_quotes = False
            cur += char
        else:
            cur += char
    tokens.append(cur)
    return tokens


def prefix_check_factory(prefix: Union[str, Iterable[str], Callable[[Client, Message], str]]):
    """
    The default message function factory.

    This provides a callable that will fire a command if the message begins with the specified
    prefix or list of prefixes.

    If ``command_prefix`` is provided to the :class:`.Client`, then it will automatically call this
    function to get a message check function to use.

    .. code-block:: python3

        # verbose form
        message_check = prefix_check_factory(["!", "?"])
        cl = Client(message_check=message_check)

        # implicit form
        cl = Client(command_prefix=["!", "?"])

    The :attr:`prefix` is set on the returned function that can be used to retrieve the prefixes
    defined to create  the function at any time.

    :param prefix: A :class:`str` or :class:`typing.Iterable[str]` that represents the prefix(es) \
        to use.
    :return: A callable that can be used for the ``message_check`` function on the client.
    """

    async def __inner(bot: Client, message: Message):
        # move prefix out of global scope
        _prefix = prefix
        matched = None

        if callable(_prefix):
            _prefix = _prefix(bot, message)
            if inspect.isawaitable(_prefix):
                _prefix = await _prefix

        if isinstance(_prefix, str):
            match = message.content.startswith(_prefix)
            if match:
                matched = _prefix

        elif isinstance(prefix, collections.Iterable):
            for i in _prefix:
                if message.content.startswith(i):
                    matched = i
                    break

        if not matched:
            return None

        tokens = split_message_content(message.content[len(matched):])
        command_word = tokens[0]

        return command_word, tokens[1:]

    __inner.prefix = prefix
    return __inner



"""
Voice connection utilities.

.. currentmodule:: curious.voice

.. autosummary::
    :toctree: voice

    voice_client
    voice_gateway
    voice_player
"""
from curious.voice.voice_client import VoiceClient
from curious.voice.voice_player import VoicePlayer



"""
A reactions-based paginator.
"""
import typing

from curious.dataclasses.channel import Channel
from curious.dataclasses.embed import Embed
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.user import User


class ReactionsPaginator(object):
    """
    A paginator for a message using reactions.
    """
    BUTTON_BACKWARDS = "◀"
    BUTTON_FORWARD   = "▶"
    BUTTON_STOP      = "⏹"

    def __init__(self, content: typing.Union[str, typing.List[str]], channel: Channel,
                 respond_to: typing.Union[Member, User], *,
                 break_at: int=2000, title: str = None):
        """
        :param content: The content to page through.
        :param channel: The channel to send the content to.
        :param respond_to: The member to respond
        :param break_at: The number of characters to break the message up into.
        :param title: The title to put above the embed.
        """
        self._content = content
        self.channel = channel
        self.respond_to = respond_to
        self.title = title

        self.bot = self.channel._bot  # hacky af

        # chunk the message up
        if isinstance(content, list):
            self._message_chunks = content
        else:
            self._message_chunks = [self._content[i:i + break_at] for i in
                                    range(0, len(self._content), break_at)]

        #: The current page this paginator is on.
        self.page = 0

        #: The message object that is being edited.
        self._message = None  # type: Message

    @classmethod
    async def paginate_response(cls, content: str,
                                responding_to: Message, *args, **kwargs) -> 'ReactionsPaginator':
        """
        Paginates a response to a message.

        :param content: The content to paginate.
        :param responding_to: The message object that you are responding to.
        """
        obb = cls(content, responding_to.channel, responding_to.author, *args, **kwargs)
        await obb.paginate()
        return obb

    async def send_current_page(self):
        """
        Sends the current page to the channel.
        """
        embed = Embed(description=self._message_chunks[self.page])
        embed.set_footer(text="Page {}/{}".format(self.page + 1, len(self._message_chunks)))

        if self._message is None:
            self._message = await self.channel.send(
                content=self.title,
                embed=embed
            )
        else:
            await self._message.edit(
                new_content=self.title,
                embed=embed
            )

    async def _add_initial_reactions(self):
        """
        Adds the initial reactions to this message.
        """
        await self._message.react(self.BUTTON_BACKWARDS)
        await self._message.react(self.BUTTON_FORWARD)
        await self._message.react(self.BUTTON_STOP)

    async def paginate(self):
        """
        Starts paginating this message.

        This will continuously listen for reactions on this message until the STOP button is pressed.
        """
        await self.send_current_page()
        await self._add_initial_reactions()

        def predicate(message: Message, author: Member, reaction: Reaction):
            """
            Inner predicate used in `wait_for`.
            """
            if message.id != self._message.id:
                return False

            if author.id != self.respond_to.id:
                return False

            return True

        # Enter the loop.
        while True:
            *_, reaction = await self.bot.wait_for("message_reaction_add", predicate=predicate)
            if reaction.emoji == self.BUTTON_FORWARD:
                if self.page < len(self._message_chunks) - 1:
                    self.page += 1
                else:
                    self.page = 0
                await self.send_current_page()

            if reaction.emoji == self.BUTTON_BACKWARDS:
                if self.page > 0:
                    self.page -= 1
                else:
                    self.page = len(self._message_chunks) - 1

                await self.send_current_page()

            if reaction.emoji == self.BUTTON_STOP:
                # remove all reactions were done here
                await self._message.remove_all_reactions()
                break

            await self._message.unreact(reaction.emoji, self.respond_to)



"""
Special helpers for events.
"""
import inspect
import logging
import typing

import curio
from multidict import MultiDict

from curious.core import client as md_client
from curious.util import remove_from_multidict

logger = logging.getLogger("curious.events")


class ListenerExit(Exception):
    """
    Raised when a temporary listener is to be exited.

    .. code-block:: python3

        def listener(ctx, message):
            if message.author.id == message.guild.owner_id:
                raise ListenerExit

    """


class EventManager(object):
    """
    A manager for events.

    This deals with firing of events and temporary listeners.
    """

    def __init__(self):
        #: A list of event hooks.
        self.event_hooks = []

        #: A MultiDict of event listeners.
        self.event_listeners = MultiDict()

        #: A MultiDict of temporary listeners.
        self.temporary_listeners = MultiDict()

    # add or removal functions
    # Events
    def add_event(self, func, name: str = None):
        """
        Add an event to the internal registry of events.

        :param name: The event name to register under.
        :param func: The function to add.
        """
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event must be a coroutine function")

        if name is None:
            evs = func.events
        else:
            evs = [name]

        for ev_name in evs:
            logger.debug("Registered event `{}` handling `{}`".format(func, name))
            self.event_listeners.add(ev_name, func)

    def remove_event(self, name: str, func):
        """
        Removes a function event.

        :param name: The name the event is registered under.
        :param func: The function to remove.
        """
        self.event_listeners = remove_from_multidict(self.event_listeners, key=name, item=func)

    def add_temporary_listener(self, name: str, listener):
        """
        Adds a new temporary listener.

        To remove the listener, you can raise ListenerExit which will exit it and remove the
        listener from the list.

        :param name: The name of the event to listen to.
        :param listener: The listener function.
        """
        self.temporary_listeners.add(name, listener)

    def remove_listener_early(self, name: str, listener):
        """
        Removes a temporary listener early.

        :param name: The name of the event the listener is registered under.
        :param listener: The listener function.
        """
        self.event_listeners = remove_from_multidict(self.event_listeners, key=name, item=listener)

    # wrapper functions
    async def _safety_wrapper(self, func, *args, **kwargs):
        """
        Ensures a coro's error is caught and doesn't balloon out.
        """
        try:
            await func(*args, **kwargs)
        except Exception as e:
            logger.exception("Unhandled exception in {}!".format(func.__name__), exc_info=True)

    async def _listener_wrapper(self, key: str, func, *args, **kwargs):
        """
        Wraps a listener, ensuring ListenerExit is handled properly.
        """
        try:
            await func(*args, **kwargs)
        except ListenerExit:
            # remove the function
            self.temporary_listeners = remove_from_multidict(self.temporary_listeners, key, func)
        except Exception:
            logger.exception("Unhandled exception in listener {}!".format(func.__name__),
                             exc_info=True)
            self.temporary_listeners = remove_from_multidict(self.temporary_listeners, key, func)

    async def wait_for(self, event_name: str, predicate = None):
        """
        Waits for an event.

        Returning a truthy value from the predicate will cause it to exit and return.

        :param event_name: The name of the event.
        :param predicate: The predicate to use to check for the event.
        """
        p = curio.Promise()
        errored = False

        async def listener(ctx, *args):
            # exit immediately if the predicate is none
            if predicate is None:
                await p.set(args)
                raise ListenerExit

            try:
                res = predicate(*args)
                if inspect.isawaitable(res):
                    res = await res
            except ListenerExit:
                # ???
                await p.set(args)
                raise
            except Exception as e:
                # something bad happened, set exception and exit
                logger.exception("Exception in wait_for predicate!")
                # signal that an error happened
                nonlocal errored
                errored = True
                await p.set(e)
                raise ListenerExit
            else:
                # exit now if result is true
                if res is True:
                    await p.set(args)
                    raise ListenerExit

        self.add_temporary_listener(name=event_name, listener=listener)
        output = await p.get()
        if errored:
            raise output

        # unwrap tuples, if applicable
        if len(output) == 1:
            return output[0]
        return output

    async def fire_event(self, event_name: str, *args, **kwargs):
        """
        Fires an event.

        :param event_name: The name of the event to fire.
        """
        if "ctx" not in kwargs:
            gateway = kwargs.pop("gateway")
            client = kwargs.pop("client")
            ctx = EventContext(client, gateway.shard_id, event_name)
        else:
            ctx = kwargs.pop("ctx")

        # always ensure hooks are ran first
        for hook in self.event_hooks:
            await curio.spawn(hook(ctx, *args, **kwargs), daemon=True)

        for handler in self.event_listeners.getall(event_name, []):
            coro = self._safety_wrapper(handler, ctx, *args, **kwargs)
            await curio.spawn(coro, daemon=True)

        for listener in self.temporary_listeners.getall(event_name, []):
            coro = self._listener_wrapper(event_name, listener, ctx, *args, **kwargs)
            await curio.spawn(coro, daemon=True)


def event(name, scan: bool = True):
    """
    Marks a function as an event.

    :param name: The name of the event.
    :param scan: Should this event be handled in scans too?
    """

    def __innr(f):
        if not hasattr(f, "events"):
            f.events = set()

        f.is_event = True
        f.events.add(name)
        f.scan = scan
        return f

    return __innr


class EventContext(object):
    """
    Represents a special context that are passed to events.
    """

    def __init__(self, cl: 'md_client.Client', shard_id: int,
                 event_name: str):
        #: The :class:`.Client` instance that this event was fired under.
        self.bot = cl

        #: The shard this event was received on.
        self.shard_id: int = shard_id
        #: The shard for this bot.
        self.shard_count: int = cl.shard_count

        #: The event name for this event.
        self.event_name: str = event_name

    @property
    def handlers(self) -> typing.List[typing.Callable[['EventContext'], None]]:
        """
        :return: A list of handlers registered for this event. 
        """
        return self.bot.events.getall(self.event_name, [])

    def change_status(self, *args, **kwargs):
        """
        Changes the current status for this shard.
        
        This takes the same arguments as :class:`~.Client.change_status`, but ignoring the shard ID.
        """
        kwargs["shard_id"] = self.shard_id
        return self.bot.change_status(*args, **kwargs)

    @property
    def gateway(self):
        """
        :return: The :class:`~.Gateway` that produced this event. 
        """
        return self.bot.gateways[self.shard_id]



"""
Wrappers for Guild objects.

.. currentmodule:: curious.dataclasses.guild
"""
import collections
import copy
import datetime
import enum
import typing
from math import ceil
from types import MappingProxyType

import curio

from curious.dataclasses import channel, emoji as dt_emoji, invite as dt_invite, \
    member as dt_member, permissions as dt_permissions, role, search as dt_search, user as dt_user, \
    voice_state as dt_vs, webhook as dt_webhook
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.presence import Presence, Status
from curious.exc import CuriousError, HTTPException, HierarchyError, PermissionsError
from curious.util import AsyncIteratorWrapper, base64ify

try:
    from curious.voice import voice_client
except ImportError:
    voice_client = None


class MFALevel(enum.IntEnum):
    """
    Represents the MFA level of a :class:`~.Guild`.
    """
    #: Used when MFA authentication is **disabled**.
    #: This means moderation actions will not require multi-factor auth.
    DISABLED = 0

    #: Used when MFA authentication is **enabled**.
    #: This means moderation actions *will* require multi-factor auth.
    ENABLED = 1


class VerificationLevel(enum.IntEnum):
    """
    Represents the verification levels for a :class:`~.Guild`.
    """
    #: No verification level.
    #: All users can speak after joining immediately.
    NONE = 0

    #: Low verification level.
    #: Users must have a verified email on their account.
    LOW = 1

    #: Medium verification level.
    #: Users must have been on Discord for longer than 5 minutes.
    MEDIUM = 2

    #: High/tableflip verification level.
    #: Users must have been on the server for longer than 10 minutes.
    TABLEFLIP = 3

    #: Extreme/double tableflip verification level.
    #: Users must have a phone number associated with their account.
    DOUBLE_TABLEFLIP = 4

    def can_speak(self, member: 'dt_member.Member') -> bool:
        """
        Checks if a :class:`~.Member` can speak in their :class:`~.Guild`.
        
        :param member: The member to check.
        :return: True if they can speak, False if they can't.
        """
        # none always allows people to speak
        if self is VerificationLevel.NONE:
            return True

        if self is VerificationLevel.LOW:
            # can't validate, assume True
            if member.user.verified is None:
                return True

            return member.user.verified is True

        if self is VerificationLevel.MEDIUM:
            dt = datetime.datetime.now() - datetime.timedelta(minutes=5)

            # ensure their created at time is before 5 minutes before now
            if member.user.created_at < dt:
                return True

        if self is VerificationLevel.TABLEFLIP:
            dt = datetime.datetime.now() - datetime.timedelta(minutes=10)

            # ensure their joined at time is before 10 minutes before now
            if member.joined_at < dt:
                return True

        # other verification levels ???
        return True


class NotificationLevel(enum.IntEnum):
    """
    Represents the default notification level for a :class:`~.Guild`.
    """
    #: All messages notify members, by default.
    ALL_MESSAGES = 0

    #: Only mentions notify members, by default.
    ONLY_MENTIONS = 1


class ContentFilterLevel(enum.IntEnum):
    """
    Represents the content filter level for a :class:`~.Guild`. 
    """
    #: No messages will be scanned.
    SCAN_NONE = 0

    #: Messages from users without roles will be scanned.
    SCAN_WITHOUT_ROLES = 1

    #: All messages will be scanned.
    SCAN_ALL = 2


class GuildChannelWrapper(collections.Mapping, collections.Iterable):
    """
    A wrapper for channels on a guild. This provides some convenience methods which make channel
    management more fluent.
    """

    def __init__(self, guild: 'Guild',
                 channels: 'typing.MutableMapping[int, channel.Channel]'):
        """
        :param guild: The :class:`~.Guild` object that owns this wrapper.
        :param channels: The dictionary of channels that this wrapper contains.
        """
        self._guild = guild
        self._channels = channels

    @property
    def view(self) -> 'typing.Mapping[int, channel.Channel]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._channels)

    def __iter__(self):
        return iter(self.view.keys())

    def __getitem__(self, key):
        return self._channels[key]

    def __len__(self):
        return len(self._channels)

    def __repr__(self):
        return "<GuildChannelWrapper channels={}>".format(self._channels)

    async def create(self, name: str, type_: 'channel.ChannelType' = None,
                     permission_overwrites: 'typing.List[dt_permissions.Overwrite]' = None,
                     *,
                     parent: 'channel.Channel' = None,
                     bitrate: int = 64, user_limit: int = 0,
                     topic: str = None) -> 'channel.Channel':
        """
        Creates a new channel.

        :param name: The name of the channel.
        :param type_: The :class:`.ChannelType` of the channel.
        :param permission_overwrites: The list of permission overwrites to use for this channel.

        For guild channels:

        :param parent: The parent :class:`.Channel` for this channel.

        For voice channels:

        :param bitrate: The bitrate of the channel, if it is a voice channel.
        :param user_limit: The maximum number of users that can be in the channel.

        For text channels:

        :param topic: The topic of the channel, or None to set no topic.
        """
        if not self._guild.me.guild_permissions.manage_channels:
            raise PermissionsError("manage_channels")

        if type_ is None:
            type_ = channel.ChannelType.TEXT

        kwargs = {
            "name": name,
            "type": type_.value,
            "permission_overwrites": permission_overwrites,
        }
        if type_ is channel.ChannelType.VOICE:
            kwargs["bitrate"] = bitrate
            kwargs["user_limit"] = user_limit

        if parent is not None:
            if parent.type != channel.ChannelType.CATEGORY:
                raise CuriousError("Cannot create channel with non-category parent")

            if type_.value == channel.ChannelType.CATEGORY:
                raise CuriousError("Cannot create category channel with category")

            kwargs["parent_id"] = parent.id

        # create a listener so we wait for the WS before editing
        async def _listener(channel: channel.Channel):
            if channel.name == name and channel.guild == self._guild:
                return True

            return False

        listener = await curio.spawn(self._guild._bot.wait_for("member_update", _listener))
        try:
            channel_data = await self._guild._bot.http.create_channel(self._guild.id, **kwargs)
            # if it's a text channel and the topic was provided, automatically add it
            if type is channel.ChannelType.TEXT and topic is not None:
                await self._guild._bot.http.edit_channel(channel_id=channel_data["id"], topic=topic)
        except:
            await listener.cancel()
            raise

        # wait on the listener
        await listener.wait()
        # we can safely assume this exists
        return self._channels[int(channel_data.get("id"))]

    async def edit(self, channel: 'channel.Channel', **kwargs):
        """
        Edits a channel.
        """
        if channel.id not in self._channels:
            raise CuriousError("This channel is not part of this guild")

        return await channel.edit(**kwargs)

    async def delete(self, channel: 'channel.Channel'):
        """
        Deletes a channel.
        """
        if channel.id not in self._channels:
            raise CuriousError("This channel is not part of this guild")

        return await channel.delete()


class GuildRoleWrapper(collections.Mapping, collections.Iterable):
    """
    A wrapper for roles on a guild. Contains some convenience methods that make role management
    more fluent.
    """

    def __init__(self, guild: 'Guild',
                 roles: 'typing.MutableMapping[int, role.Role]'):
        """
        :param guild: The :class:`~.Guild` object that owns this wrapper.
        :param roles: The dictionary of roles that this wrapper contains.
        """
        self._guild = guild
        self._roles = roles

    @property
    def view(self) -> 'typing.Mapping[int, role.Role]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._roles)

    def __iter__(self):
        return iter(self.view.keys())

    def __getitem__(self, key):
        return self._roles[key]

    def __len__(self):
        return len(self._roles)

    def __repr__(self):
        return "<GuildRoleWrapper roles={}>".format(self._roles)

    async def create(self, **kwargs) -> 'role.Role':
        """
        Creates a new role in this guild.

        :return: A new :class:`~.Role`.
        """
        if not self._guild.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        role_obb = role.Role(client=self._guild._bot,
                             **(await self._guild._bot.http.create_role(self._guild.id)))
        self._roles[role_obb.id] = role_obb
        role_obb.guild_id = self._guild.id
        return await role_obb.edit(**kwargs)

    def edit(self, role: 'role.Role', **kwargs):
        """
        Edits a role.
        """
        if role.id not in self._roles:
            raise CuriousError("This role is not part of this guild")

        return role.edit(**kwargs)

    def delete(self, role: 'role.Role'):
        """
        Deletes a role.
        """
        if role.id not in self._roles:
            raise CuriousError("This role is not part of this guild")

        return role.delete()


class GuildEmojiWrapper(collections.Mapping, collections.Iterable):
    """
    Wrapper for emoji objects for a guild.
    """

    def __init__(self, guild: 'Guild',
                 emojis: 'typing.MutableMapping[int, dt_emoji.Emoji]'):
        """
        :param guild: The :class:`.Guild` object that owns this wrapper.
        :param emojis: The dictionary of emojis that this wrapper contains.
        """
        self._guild = guild
        self._emojis = emojis

    @property
    def view(self) -> 'typing.Mapping[int, dt_emoji.Emoji]':
        """
        :return: A read-only view into the channels for this wrapper.
        """
        return MappingProxyType(self._emojis)

    def __iter__(self):
        return iter(self._emojis.keys())

    def __getitem__(self, key):
        return self._emojis[key]

    def __len__(self):
        return len(self._emojis)

    def __repr__(self):
        return "<GuildEmojiWrapper emojis={}>".format(self._emojis)

    async def create(self, *,
                     name: str, image_data: typing.Union[str, bytes],
                     roles: 'typing.List[role.Role]' = None) -> 'dt_emoji.Emoji':
        """
        Creates a new emoji in this guild.

        :param name: The name of the emoji.
        :param image_data: The bytes image data or the str base64 data for the emoji.
        :param roles: A list of roles this emoji is locked to.
        :return: The :class:`.Emoji` created.
        """
        if isinstance(image_data, bytes):
            image_data = base64ify(image_data)

        if roles is not None:
            roles = [r.id for r in roles]

        emoji_data = await self._guild._bot.http.create_guild_emoji(self._guild.id,
                                                                    name=name,
                                                                    image_data=image_data,
                                                                    roles=roles)
        emoji = dt_emoji.Emoji(**emoji_data, client=self._guild._bot)
        return emoji


class Guild(Dataclass):
    """
    Represents a guild object on Discord.
    """

    __slots__ = (
        "id", "unavailable", "name", "afk_timeout", "region",
        "mfa_level", "verification_level", "notification_level", "content_filter_level", "features",
        "shard_id", "_roles", "_members", "_channels", "_emojis", "member_count", "_voice_states",
        "_large", "_chunks_left", "_finished_chunking", "_icon_hash", "_splash_hash",
        "owner_id", "afk_channel_id", "system_channel_id",
        "voice_client",
        "channels", "roles", "emojis"
    )

    def __init__(self, bot, **kwargs):
        """
        Creates a new Guild object.
        """
        super().__init__(kwargs.get("id"), bot)

        #: If the guild is unavailable or not.
        #: If this is True, many fields return `None`.
        self.unavailable = kwargs.get("unavailable", False)

        # Placeholder values.
        #: The name of this guild.
        self.name = None  # type: str
        #: The icon hash of this guild.
        #: Used to construct the icon URL later.
        self._icon_hash = None  # type: str
        #: The splash hash of this guild.
        #: Used to construct the splash URL later.
        self._splash_hash = None  # type: str

        #: The AFK channel ID of this guild.
        self.afk_channel_id = None  # type: int

        #: The ID of the system channel for this guild.
        #: This is where welcome messages and the likes are sent.
        #: Effective replacement for default channel for bots.
        self.system_channel_id = None

        #: The owner ID of this guild.
        self.owner_id = None  # type: int

        #: The AFK timeout for this guild.
        self.afk_timeout = None  # type: int

        #: The voice region of this guild.
        self.region = None  # type: str

        #: The features this guild has.
        self.features = None  # type: typing.List[str]

        #: The MFA level of this guild.
        self.mfa_level = MFALevel.DISABLED  # type: MFALevel
        #: The verification level of this guild.
        self.verification_level = VerificationLevel.NONE  # type: int
        #: The notification level of this guild.
        self.notification_level = NotificationLevel.ALL_MESSAGES
        #: The content filter level of this guild.
        self.content_filter_level = ContentFilterLevel.SCAN_NONE

        #: The shard ID this guild is associated with.
        self.shard_id = None

        #: The roles that this guild has.
        self._roles = {}
        #: The members of this guild.
        self._members = {}
        #: The channels of this guild.
        self._channels = {}
        #: The emojis that this guild has.
        self._emojis = {}
        #: The voice states that this guild has.
        self._voice_states = {}

        #: The number of numbers this guild has.
        #: This is automatically updated.
        self.member_count = 0  # type: int

        #: Is this guild a large guild?
        self._large = None  # type: bool

        #: Has this guild finished chunking?
        self._finished_chunking = curio.Event()
        self._chunks_left = 0

        #: The current voice client associated with this guild.
        self.voice_client = None

        #: The :class:`.GuildChannelWrapper` that wraps the channels in this Guild.
        self.channels = GuildChannelWrapper(self, self._channels)
        #: The :class:`.GuildRoleWrapper` that wraps the roles in this Guild.
        self.roles = GuildRoleWrapper(self, self._roles)
        #: The :class:`.GuildEmojiWrapper` that wraps the emojis in this Guild.
        self.emojis = GuildEmojiWrapper(self, self._emojis)

        if kwargs:
            self.from_guild_create(**kwargs)

    def _copy(self):
        return copy.copy(self)

    def __repr__(self):
        return "<Guild id='{}' name='{}' members='{}'>".format(self.id, self.name,
                                                               self.member_count)

    def __str__(self):
        return repr(self)

    @property
    def members(self) -> 'typing.Mapping[int, dt_member.Member]':
        """
        :return: A mapping of :class:`~.Member` that represent members on this guild.
        """
        return MappingProxyType(self._members)

    @property
    def owner(self) -> 'dt_member.Member':
        """
        :return: A :class:`~.Member` object that represents the owner of this guild.
        """
        return self._members[self.owner_id]

    @property
    def me(self) -> 'typing.Union[dt_member.Member, None]':
        """
        :return: A :class:`~.Member` object that represents the current user in this guild.
        """
        try:
            return self._members[self._bot.user.id]
        except KeyError:
            return None

    @property
    def default_role(self) -> 'typing.Union[role.Role, None]':
        """
        :return: A :class:`~.Role` that represents the default role of this guild.
        """
        try:
            return self.roles[self.id]
        except KeyError:
            return None

    @property
    def system_channel(self) -> 'typing.Union[channel.Channel, None]':
        """
        :return: A :class:`~.Channel` that represents the system channel for this guild.
        """
        try:
            return self.channels[self.system_channel_id]
        except KeyError:
            return None

    @property
    def afk_channel(self) -> 'typing.Union[channel.Channel, None]':
        """
        :return: A :class:`~.Channel` representing the AFK channel for this guild.
        """
        try:
            return self._channels[self.afk_channel_id]
        except IndexError:
            # the afk channel CAN be None
            return None

    @property
    def embed_url(self) -> str:
        """
        Gets the default embed url for this guild.
        If the widget is not enabled, this endpoint will 404.
        
        :return: The embed URL for this guild. 
        """
        return (self._bot.http.GUILD_BASE + "/embed.png").format(guild_id=self.id)

    # for parity with inviteguild
    @property
    def presence_count(self) -> int:
        """
        :return: The number of members with a non-Invisible presence. 
        """
        return sum(1 for member in self._members.values() if member.status is not Status.OFFLINE)

    # Presence methods
    def members_with_status(self,
                            status: Status) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        A generator that returns the members that match the specified status.
        """
        for member in self.members.values():
            if member.status == status:
                yield member

    @property
    def online_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of online :class:`~.Member` objects.
        """
        return self.members_with_status(Status.ONLINE)

    @property
    def idle_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of idle :class:`~.Member` objects.
        """
        return self.members_with_status(Status.IDLE)

    @property
    def dnd_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of DnD :class:`~.Member` objects.
        """
        return self.members_with_status(Status.DND)

    @property
    def offline_members(self) -> 'typing.Generator[dt_member.Member, None, None]':
        """
        :return: A generator of offline/invisible :class:`~.Member` objects.
        """
        return self.members_with_status(Status.OFFLINE)

    @property
    def search(self) -> 'dt_search.SearchQuery':
        """
        :return: A :class:`~.SearchQuery` that can be used to search this guild's messages. 
        """
        return dt_search.SearchQuery(guild=self)

    def get_embed_url(self, *, style: str = "banner1") -> str:
        """
        Gets an embed URL for this guild in a specified style.
        
        :param style: The style to get. 
        :return: The embed URL for this guild.
        """
        valid_styles = ('banner1', 'banner3', 'banner2', 'shield', 'banner4')
        if style not in valid_styles:
            raise ValueError("Style must be in {}".format(valid_styles))

        return self.embed_url + "?style={}".format(style)

    def find_member(self, search_str: str) -> 'dt_member.Member':
        """
        Attempts to find a member in this guild by name#discrim.
        This will also search nicknames.

        The discriminator is optional, but if provided allows better matching.

        :param search_str: The name#discrim pair to search for.
        :return: A :class:`~.Member` object that represents the member, or None if no member could \ 
            be found.
        """
        sp = search_str.rsplit("#", 1)
        if len(sp) == 1:
            # Member name only :(
            predicate = lambda member: member.user.name == sp[0] or member.nickname == sp[0]
        else:
            # Discriminator too!
            # Don't check nicknames for this.
            predicate = lambda member: member.user.name == sp[0] \
                                       and member.user.discriminator == sp[1]

        filtered = filter(predicate, self.members.values())
        return next(filtered, None)

    # creation methods
    def start_chunking(self) -> None:
        """
        Marks a guild to start guild chunking.
        
        This will clear the chunking event, and calculate the number of member chunks required.
        """
        self._finished_chunking.clear()
        self._chunks_left = ceil(self.member_count / 1000)

    async def wait_until_chunked(self) -> None:
        """
        Waits until the guild has finished chunking.

        Useful for when you join a big guild.
        """
        await self._finished_chunking.wait()

    def _handle_member_chunk(self, members: list):
        """
        Handles a chunk of members.
        
        :param members: A list of member data dictionaries as returned from Discord.
        """
        if self._chunks_left >= 1:
            # We have a new chunk, so decrement the number left.
            self._chunks_left -= 1

        for member_data in members:
            id = int(member_data["user"]["id"])
            if id in self._members:
                member_obj = self._members[id]
            else:
                member_obj = dt_member.Member(self._bot, **member_data)

            member_obj.nickname = member_data.get("nick", member_obj.nickname)
            member_obj.guild_id = self.id

            self._members[member_obj.id] = member_obj

    def _handle_emojis(self, emojis: list):
        """
        Handles the emojis for this guild.
        
        :param emojis: A list of emoji objects from Discord.
        """
        for emoji in emojis:
            emoji_obj = dt_emoji.Emoji(**emoji, client=self._bot)
            emoji_obj.guild_id = self
            self._emojis[emoji_obj.id] = emoji_obj

    def from_guild_create(self, **data: dict) -> 'Guild':
        """
        Populates the fields from a GUILD_CREATE event.

        :param data: The GUILD_CREATE data to use.
        """
        self.unavailable = data.get("unavailable", False)

        if self.unavailable:
            # We can't use any of the extra data here, so don't bother.
            return self

        self.name = data.get("name")  # type: str
        self._icon_hash = data.get("icon")  # type: str
        self._splash_hash = data.get("splash")  # type: str
        self.owner_id = int(data.get("owner_id"))  # type: int
        self._large = data.get("large", None)
        self.features = data.get("features", [])
        self.region = data.get("region")
        afk_channel_id = data.get("afk_channel_id")
        if afk_channel_id is not None:
            afk_channel_id = int(afk_channel_id)
        self.afk_channel_id = afk_channel_id
        self.afk_timeout = data.get("afk_timeout")

        self.verification_level = VerificationLevel(data.get("verification_level", 0))
        self.mfa_level = MFALevel(data.get("mfa_level", 0))
        self.notification_level = NotificationLevel(data.get("default_message_notifications"))
        self.content_filter_level = ContentFilterLevel(data.get("explicit_content_filter", 0))

        self.member_count = data.get("member_count", 0)

        # Create all the Role objects for the server.
        for role_data in data.get("roles", []):
            role_obj = role.Role(self._bot, **role_data)
            role_obj.guild_id = self.id
            self._roles[role_obj.id] = role_obj

        # Create all the Member objects for the server.
        self._handle_member_chunk(data.get("members", []))

        for presence in data.get("presences", []):
            member_id = int(presence["user"]["id"])
            member_obj = self._members.get(member_id)

            if not member_obj:
                continue

            member_obj.presence = Presence(**presence)

        # Create all of the channel objects.
        for channel_data in data.get("channels", []):
            channel_obj = channel.Channel(self._bot, **channel_data)
            channel_obj.guild_id = self.id
            channel_obj._update_overwrites(channel_data.get("permission_overwrites", []),
                                           guild=self)
            self._channels[channel_obj.id] = channel_obj

        # Create all of the voice states.
        for vs_data in data.get("voice_states", []):
            user_id = int(vs_data.get("user_id", 0))
            member = self.members.get(user_id)
            if not member:
                # o well
                continue

            voice_state = dt_vs.VoiceState(**vs_data, client=self._bot)

            vs_channel = self._channels.get(int(vs_data.get("channel_id", 0)))
            if vs_channel is not None:
                voice_state.channel_id = vs_channel.id
                voice_state.guild_id = self.id

            self._voice_states[voice_state.user_id] = voice_state

        # Create all of the emoji objects for the server.
        self._handle_emojis(data.get("emojis", []))

    @property
    def large(self) -> bool:
        """
        :return: If this guild is large or not (>= 250 members).
        """
        if self._large is not None:
            return self._large

        return self.member_count >= self._bot.gateways[self.shard_id].large_threshold

    @property
    def bans(self) -> 'typing.AsyncIterator[dt_user.User]':
        """
        :return: A :class:`~.AsyncIteratorWrapper` that yields :class:`~.User` objects that are
            banned.
        """
        return AsyncIteratorWrapper(self.get_bans())

    @property
    def invites(self) -> 'typing.AsyncIterator[dt_invite.Invite]':
        """
        :return: A class:`~.AsyncIteratorWrapper` that yields :class:`~.Invite` objects for this
            guild.
        """
        return AsyncIteratorWrapper(self.get_invites())

    @property
    def icon_url(self) -> str:
        """
        :return: The icon URL for this guild, or None if one isn't set.
        """
        if self._icon_hash:
            return "https://cdn.discordapp.com/icons/{}/{}.webp".format(self.id, self._icon_hash)

    @property
    def splash_url(self) -> str:
        """
        :return: The splash URL for this guild, or None if one isn't set.
        """
        if self._splash_hash:
            return "https://cdn.discordapp.com/splashes/{}/{}.webp".format(self.id,
                                                                           self._splash_hash)

    # Guild methods.
    async def leave(self) -> None:
        """
        Leaves the guild.
        """
        await self._bot.http.leave_guild(self.id)

    async def connect_to_voice(self, channel: 'channel.Channel') -> 'voice_client.VoiceClient':
        """
        Connects to a voice channel in this guild.

        :param channel: The :class:`~.Channel` to connect to.
        :return: The :class:`VoiceClient` that was connected to this guild.
        """
        if voice_client is None:
            raise RuntimeError("Cannot to voice - voice support is not installed")

        if channel.guild != self:
            raise CuriousError("Cannot use channel from a different guild")

        if self.voice_client is not None and self.voice_client.open:
            raise CuriousError("Voice client already exists in this guild")

        gw = self._bot._gateways[self.shard_id]
        self.voice_client = await voice_client.VoiceClient.create(self._bot, gw, channel)
        await self.voice_client.connect()
        return self.voice_client

    async def get_invites(self) -> 'typing.List[dt_invite.Invite]':
        """
        Gets the invites for this guild.
        :return: A list :class:`~.Invite` objects.
        """
        invites = await self._bot.http.get_invites_for(self.id)
        invites = [dt_invite.Invite(self._bot, **i) for i in invites]

        try:
            invite = await self.get_vanity_invite()
        except (CuriousError, HTTPException):
            pass
        else:
            if invite is not None:
                invites.append(invite)

        return invites

    async def get_bans(self) -> 'typing.List[dt_user.User]':
        """
        Gets the bans for this guild.
        :return: A list of User objects, one for each ban.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        bans = await self._bot.http.get_bans(self.id)
        users = []

        for user_data in bans:
            # TODO: Audit log stuff, if it ever comes out.
            user_data = user_data.get("user", None)
            users.append(dt_user.User(self._bot, **user_data))

        return users

    async def kick(self, victim: 'dt_member.Member'):
        """
        Kicks somebody from the guild.

        :param victim: The :class:`~.Member` to kick.
        """
        if not self.me.guild_permissions.kick_members:
            raise PermissionsError("kick_members")

        if victim.guild != self:
            raise ValueError("Member must be from this guild (try `member.user` instead)")

        if victim.top_role >= self.me.top_role:
            raise HierarchyError("Top role is equal to or lower than victim's top role")

        victim_id = victim.user.id

        await self._bot.http.kick_member(self.id, victim_id)

    async def ban(self, victim: 'typing.Union[dt_member.Member, dt_user.User]', *,
                  delete_message_days: int = 7):
        """
        Bans somebody from the guild.

        This can either ban a :class:`~.Member`, in which they must be in the guild. 
        Or this can ban a :class:`~.User`, which does not need to be in the guild.

        Example for banning a member:

        .. code:: python

            member = guild.members[66237334693085184]
            await guild.ban(member)

        Example for banning a user:

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.ban(user)

        :param victim: The :class:`~.Member` or :class:`~.User` object to ban.
        :param delete_message_days: The number of days to delete messages.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        if isinstance(victim, dt_member.Member):
            if self.owner == victim:
                raise HierarchyError("Cannot ban the owner")

            if victim.guild != self:
                raise ValueError("Member must be from this guild (try `member.user` instead)")

            if victim.top_role >= self.me.top_role:
                raise HierarchyError("Top role is equal to or lower than victim's top role")

            victim_id = victim.user.id

        elif isinstance(victim, dt_user.User):
            victim_id = victim.id

        else:
            raise TypeError("Victim must be a Member or a User")

        await self._bot.http.ban_user(guild_id=self.id, user_id=victim_id,
                                      delete_message_days=delete_message_days)

    async def unban(self, user: 'dt_user.User'):
        """
        Unbans a user from this guild.

        Example for unbanning the first banned user:

        .. code:: python

            user = next(await guild.get_bans())
            await guild.unban(user)

        To unban an arbitrary user, use :meth:`~.Client.get_user`.

        .. code:: python

            user = await client.get_user(66237334693085184)
            await guild.unban(user)

        :param user: The :class:`~.User` to forgive and unban.
        """
        if not self.me.guild_permissions.ban_members:
            raise PermissionsError("ban_members")

        forgiven_id = user.id

        await self._bot.http.unban_user(self.id, forgiven_id)

    async def get_webhooks(self) -> 'typing.List[dt_webhook.Webhook]':
        """
        Gets the webhooks for this guild.

        :return: A list of :class:`~.Webhook` objects for the guild.
        """
        webhooks = await self._bot.http.get_webhooks_for_guild(self.id)
        obbs = []

        for webhook in webhooks:
            obbs.append(self._bot.state.make_webhook(webhook))

        return obbs

    async def delete_webhook(self, webhook: 'dt_webhook.Webhook'):
        """
        Deletes a webhook in this guild.

        :param webhook: The :class:`~.Webhook` to delete.
        """
        if not self.me.guild_permissions.manage_webhooks:
            raise PermissionsError("manage_webhooks")

        await self._bot.http.delete_webhook(webhook.id)

    async def add_roles(self, member: 'dt_member.Member', *roles: 'role.Role') \
            -> 'dt_member.Member':
        """
        Adds roles to a member.

        This will wait until the gateway returns the GUILD_MEMBER_UPDATE with the new role list 
        for the member before returning.

        .. code:: python

            roles = filter(lambda r: "Mod" in r.name, guild.roles)
            await guild.add_roles(member, *roles)

        :param member: The :class:`~.Member` to add roles to.
        :param roles: An iterable of :class:`~.Role` objects to add.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        # Ensure we can add all of these roles.
        for _r in roles:
            if _r >= self.me.top_role:
                msg = "Cannot add role {} - it has a higher or equal position to our top role" \
                    .format(_r.name)
                raise HierarchyError(msg)

        async def _listener(before, after: dt_member.Member):
            if after.id != member.id:
                return False

            if not all(role in after.roles for role in roles):
                return False

            return True

        listener = await curio.spawn(self._bot.wait_for("member_update", _listener))
        role_ids = set([_r.id for _r in member.roles] + [_r.id for _r in roles])

        try:
            await self._bot.http.edit_member_roles(self.id, member.id, role_ids)
        except:
            await listener.cancel()
            raise

        # Now wait for the event to happen on the gateway.
        await listener.join()

        return member

    async def remove_roles(self, member: 'dt_member.Member', *roles: 'role.Role'):
        """
        Removes roles from a member.

        This will wait until the gateway fires a GUILD_MEMBER_UPDATE.

        :param member: The :class:`~.Member` to remove roles from.
        :param roles: An iterable of :class:`Role` to remove.
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        for _r in roles:
            if _r >= self.me.top_role:
                msg = "Cannot remove role {} - it has a higher or equal position to our top role" \
                    .format(_r.name)
                raise HierarchyError(msg)

        async def _listener(before, after: dt_member.Member):
            if after.id != member.id:
                return False

            if not all(role not in after.roles for role in roles):
                return False

            return True

        # Calculate the roles to keep.
        to_keep = set(member.roles) - set(roles)

        role_ids = set([_r.id for _r in to_keep])
        listener = await curio.spawn(self._bot.wait_for("member_update", _listener))

        try:
            await self._bot.http.edit_member_roles(self.id, member.id, role_ids)
        except:
            await listener.cancel()
            raise
        await listener.join()

        return member

    async def change_nickname(self, member: 'dt_member.Member', new_nickname: str):
        """
        Changes the nickname of a member.

        :param member: The :class:`~.Member` to change the nickname of.
        :param new_nickname: The new nickname.
        """
        me = False
        if member == self.me:
            me = True
            if not self.me.guild_permissions.change_nickname:
                raise PermissionsError("change_nickname")
        else:
            if not self.me.guild_permissions.manage_nicknames:
                raise PermissionsError("manage_nicknames")

        if member.top_role >= self.me.top_role and member != self.me:
            raise HierarchyError("Top role is equal to or lower than victim's top role")

        if new_nickname is not None and len(new_nickname) > 32:
            raise ValueError("Nicknames cannot be longer than 32 characters")

        coro = self._bot.http.change_nickname(self.id, new_nickname,
                                              member_id=member.id, me=me)

        async def _listener(before, after):
            return after.guild == self and after.id == member.id

        coro = self._bot.wait_for("member_update", _listener)
        listener = await curio.spawn(coro)  # type: curio.Task

        try:
            await coro
        except:
            await listener.cancel()
            raise
        await listener.join()

        return member

    async def change_role_positions(self, roles: 'typing.Union[typing.Dict[role.Role, int], '
                                                 'typing.List[typing.Tuple[role.Role, int]]]'):
        """
        Changes the positions of a mapping of roles.

        :param roles: A dict or iterable of two-item tuples of new roles that is in the format of \
            (role, position).
        """
        if not self.me.guild_permissions.manage_roles:
            raise PermissionsError("manage_roles")

        if isinstance(roles, dict):
            roles = roles.items()

        to_send = []
        for r, new_position in roles:
            if new_position >= self.me.top_role.position:
                raise HierarchyError("Cannot move role above our top role")

            to_send.append((str(r.id), new_position))

        to_send = [(str(r.id), new_position) for (r, new_position) in roles]
        await self._bot.http.edit_role_positions(to_send)

    async def change_voice_state(self, member: 'dt_member.Member', *,
                                 deaf: bool = None, mute: bool = None):
        """
        Changes the voice state of a member.

        :param member: The :class:`~.Member` to change the voice state of.
        :param deaf: Should this member be deafened?
        :param mute: Should this member be muted?
        """
        if member.voice is None:
            raise CuriousError("Cannot change voice state of member not in voice")

        await self._bot.http.edit_member_voice_state(self.id, member.id, deaf=deaf, mute=mute)
        return member.voice

    async def modify_guild(self, *, afk_channel: 'channel.Channel' = None,
                           verification_level: VerificationLevel = None,
                           content_filter_level: ContentFilterLevel = None,
                           notification_level: NotificationLevel = None,
                           **kwargs):
        """
        Edits this guild.

        For a list of available arguments, see 
        https://discordapp.com/developers/docs/resources/guild#modify-guild.
        
        :param afk_channel: The :class:`~.Channel` that represents the AFK voice channel.
        :param verification_level: The :class:`.VerificationLevel` to use for this guild.
        :param content_filter_level: The :class:`.ContentFilterLevel` to use for this guild.
        :param notification_level: The :class:`.NotificationLevel` to use for this guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        if afk_channel is not None:
            kwargs["afk_channel_id"] = afk_channel.id

        if verification_level is not None:
            kwargs["verification_level"] = verification_level.value

        if notification_level is not None:
            kwargs["default_message_notifications"] = notification_level.value

        if content_filter_level is not None:
            kwargs["explicit_content_filter"] = content_filter_level.value

        await self._bot.http.edit_guild(self.id, **kwargs)
        return self

    async def change_icon(self, icon_content: bytes):
        """
        Changes the icon for this guild.

        :param icon_content: The bytes that represent the icon of the guild.
        """
        if not self.me.guild_permissions.manage_server:
            raise PermissionsError("manage_server")

        image = base64ify(icon_content)
        await self._bot.http.edit_guild(self.id,
                                        icon_content=image)

    def upload_icon(self, path):
        """
        Uploads a new icon for the guild.

        :param path: A path-like object to use to upload.
        """
        with open(path, 'rb') as f:
            return self.change_icon(f.read())

    async def get_widget_info(self) -> 'typing.Tuple[bool, typing.Union[None, channel.Channel]]':
        """
        Gets the widget info for the current guild.
        
        :return: A two-item tuple: If this widget is enabled, and the channel the widget has an \ 
            invite for. 
        """
        info = await self._bot.http.get_widget_status(self.id)
        return info.get("enabled", False), self.channels.get(int(info.get("channel_id", 0)))

    async def edit_widget(self, *,
                          status: bool = None, channel: 'channel.Channel' = -1):
        """
        Edits the widget for this guild.
        
        :param status: The status of this widget: True or False.
        :param channel: The channel object to set the instant invite to.
        """
        if channel is None:
            channel_id = None
        elif channel == -1:
            channel_id = 0
        else:
            channel_id = channel.id

        await self._bot.http.edit_widget(self.id, enabled=status, channel_id=channel_id)

    async def get_vanity_invite(self) -> 'typing.Union[None, dt_invite.Invite]':
        """
        Gets the vanity :class:`~.Invite` for this guild.

        :return: The :class:`~.Invite` that corresponds with this guild, if it has one.
        """
        if 'vanity-url' not in self.features:
            return None

        try:
            resp = await self._bot.http.get_vanity_url(self.id)
        except HTTPException as e:
            if e.error_code != 50020:
                raise
            else:
                return None

        code = resp.get("code", None)
        if code is None:
            return None

        invite_data = await self._bot.http.get_invite(code)
        invite = dt_invite.Invite(self._bot, **invite_data)

        return invite

    async def set_vanity_invite(self, url: str) -> 'dt_invite.Invite':
        """
        Sets the vanity :class:`~.Invite` for this guild.

        :param url: The code to use for this guild.
        :return: The :class:`~.Invite` produced.
        """
        if 'vanity-url' not in self.features:
            raise CuriousError("This guild has no vanity URL")

        try:
            resp = await self._bot.http.edit_vanity_url(self.id, url)
        except HTTPException as e:
            if e.error_code != 50020:
                raise

            raise CuriousError("This guild has no vanity URL")

        code = resp.get("code", None)
        if code is None:
            return None

        invite_data = await self._bot.http.get_invite(code)
        invite = dt_invite.Invite(self._bot, **invite_data)

        return invite

    @property
    def recent_mentions(self):
        """
        :return: A :class:`~.AsyncIteratorWrapper` that can be used to get all the mentions for \ 
            this user in this guild.
        """
        return AsyncIteratorWrapper(
            self.get_recent_mentions(limit=100, everyone_mentions=True, role_mentions=True)
        )

    def get_recent_mentions(self, *,
                            limit: int = 25,
                            everyone_mentions: bool = True, role_mentions: bool = True):
        """
        Gets mentions in this guild.
        """
        if self.me.user.bot is True:
            raise CuriousError("Cannot get recent mentions on bot accounts")

        return self.me.user.get_recent_mentions(guild=self, limit=limit,
                                                everyone_mentions=everyone_mentions,
                                                role_mentions=role_mentions)



"""
Misc utilities shared throughout the library.
"""
import base64
import collections
import datetime
import imghdr
import inspect
import typing

from multidict import MultiDict

NO_ITEM = object()


def remove_from_multidict(d: MultiDict, key: str, item: typing.Any):
    """
    Removes an item from a multidict key.
    """
    # works by popping all, removing, then re-adding into
    i = d.popall(key, [])
    if item in i:
        i.remove(item)

    for n in i:
        d.add(key, n)

    return d


class AsyncIteratorWrapper(collections.AsyncIterator):
    """
    Wraps a coroutine that returns a sequence of items into something that can iterated over
    asynchronously.
    
    .. code-block:: python3
    
        async def a():
            # ... some long op
            return [..., ..., ...]
         
        it = AsyncIteratorWrapper(a())
        
        async for item in it:
            print(item)
    
    """

    def __init__(self, coro: typing.Awaitable[typing.List[typing.Any]]):
        self.coro = coro

        self.items = collections.deque()

        self._filled = False

    async def _fill(self):
        self.items.extend(await self.coro)
        self._filled = True

    async def __anext__(self):
        if not self._filled:
            await self._fill()

        try:
            return self.items.popleft()
        except IndexError:
            raise StopAsyncIteration

    # helper methods
    async def next(self, default=NO_ITEM) -> typing.Any:
        """
        Gets the next item from this iterable.
        """
        try:
            return await self.__anext__()
        except StopAsyncIteration:
            if default == NO_ITEM:
                raise

            return default

    async def all(self) -> typing.List[typing.Any]:
        """
        Gets a flattened list of items from this iterator.
        """
        items = []

        async for item in self:
            items.append(item)

        return items


def base64ify(image_data: bytes):
    """
    Base64-ifys an image to send to discord.

    :param image_data: The data of the image to use.
    :return: A string containing the encoded image.
    """
    # Convert the avatar to base64.
    mimetype = imghdr.what(None, image_data)
    if not mimetype:
        raise ValueError("Invalid image type")

    b64_data = base64.b64encode(image_data).decode()
    return "data:{};base64,{}".format(mimetype, b64_data)


def to_datetime(timestamp: str) -> datetime.datetime:
    """
    Converts a Discord-formatted timestamp to a datetime object.

    :param timestamp: The timestamp to convert.
    :return: The :class:`datetime.datetime` object that corresponds to this datetime.
    """
    if timestamp is None:
        return

    if timestamp.endswith("+00:00"):
        timestamp = timestamp[:-6]

    try:
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
    except ValueError:
        # wonky datetimes
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")


def replace_quotes(item: str) -> str:
    """
    Replaces the quotes in a string, but only if they are un-escaped.
    
    .. code-block:: python3
    
        some_weird_string = r'"this is quoted and removed" but \" that was kept and this isn't \\"'
        replace_quotes(some_weird_string)  # 'this is quoted and removed but " that was kept but this isnt \\'

    :param item: The string to scan.
    :return: The string, with quotes replaced.
    """
    # A list is used because it can be appended easily.
    final_str_arr = []

    for n, char in enumerate(item):
        # only operate if the previous char actually exists
        if n - 1 < 0:
            if char != '"':
                final_str_arr.append(char)

            continue

        # Complex quoting rules!
        # If it's a SINGLE backslash, don't append it.
        # If it's a double backslash, append it.
        if char == '\\':
            if item[n - 1] == "\\":
                # double backslash, append it
                final_str_arr.append(char)

            continue

        if char == '"':
            # check to see if it's escaped
            if item[n - 1] == '\\':
                # if the last char on final_str_arr is NOT a backslash, we want to keep it.
                if len(final_str_arr) > 0 and final_str_arr[-1] != '\\':
                    final_str_arr.append('"')

            continue

        # None of the above were hit, so add it anyway and continue.
        final_str_arr.append(char)

    return "".join(final_str_arr)


def _traverse_stack_for(t: type):
    """
    Traverses the stack for an object of type `t`.

    :param t: The type of the object.
    :return: The object, if found.
    """
    for fr in inspect.stack():
        frame = fr.frame
        try:
            locals = frame.locals
        except AttributeError:
            # idk
            continue
        else:
            for object in locals.values():
                if type(object) is t:
                    return object
        finally:
            # prevent reference cycles
            del fr


def _ad_getattr(self, key: str):
    try:
        return self[key]
    except KeyError as e:
        raise AttributeError(key) from e


attrdict = type("attrdict", (dict,), {
    "__getattr__": _ad_getattr,
    "__setattr__": dict.__setitem__,
    "__doc__": "A dict that allows attribute access as well as item access for "
               "keys."
})



"""
Wrappers for Status objects.

.. currentmodule:: curious.dataclasses.presence
"""

import enum


class Status(enum.Enum):
    """
    Represents a Member's status.
    """
    #: Corresponds to online (green dot).
    ONLINE = 'online'

    #: Corresponds to offline (gray dot).
    OFFLINE = 'offline'

    #: Corresponds to idle (yellow dot).
    IDLE = 'idle'

    #: Corresponds to Do Not Disturb (red dot).
    DND = 'dnd'

    #: Corresponds to invisible (gray dot).
    INVISIBLE = 'invisible'

    @property
    def strength(self) -> int:
        """
        :return: The strength of the presence, when computing the final presence on multiple \ 
            connections. 
        """
        return strengths.index(self)


strengths = [Status.OFFLINE, Status.INVISIBLE, Status.IDLE, Status.DND, Status.ONLINE]


class Game(object):
    """
    Represents a game object.
    """

    __slots__ = "type", "url", "name"

    def __init__(self, **kwargs):
        #: The type of game this is.
        self.type = kwargs.get("type", 0)
        #: The stream URL this game is for.
        self.url = kwargs.get("url", None)
        #: The name of the game being played.
        self.name = kwargs.get("name", None)

    def to_dict(self) -> dict:
        """
        :return: The dict representation of this object. 
        """
        d = {
            "name": self.name,
            "type": self.type,
        }
        if self.url is not None:
            d["url"] = self.url

        return d

    def __repr__(self):
        return "<Game name='{}' type={} url={}>".format(self.name, self.type, self.url)


class Presence(object):
    """
    Represents a presence on a member.
    """

    __slots__ = "_status", "_game"

    def __init__(self, **kwargs):
        #: The :class:`~.Status` for this presence.
        self._status = None  # type: Status
        # prevent dupe code by using our setter
        self.status = kwargs.get("status", Status.OFFLINE)

        game = kwargs.get("game", None)
        #: The :class:`~.Game` for this presence.
        self._game = None  # type: Game

        # NB: this does a property set to ensure the types are right.
        self.game = game

    def __repr__(self):
        return "<Presence status={} game='{}'>".format(self.status, self.game)

    @property
    def status(self) -> Status:
        """
        :return: The :class:`~.Status` associated with this presence. 
        """
        return self._status

    @status.setter
    def status(self, value):
        if value is None:
            return

        if not isinstance(value, Status):
            value = Status(value)

        self._status = value

    @property
    def game(self) -> Game:
        """
        :return: The :class:`~.Game` associated with this presence.
        """
        return self._game

    @game.setter
    def game(self, value):
        if value is None:
            self._game = None
            return

        if not isinstance(value, Game):
            value = Game(**value)

        self._game = value

    @property
    def strength(self) -> int:
        """
        :return: The strength for this status.
        """
        return self.status.strength



"""
A special gateway client for Voice.
"""
import enum
import json
import logging
import socket
import threading
import time
import typing
import zlib

import curio
from asyncwebsockets import Websocket, WebsocketBytesMessage, WebsocketClosed, connect_websocket
from curio.thread import AWAIT, async_thread

from curious.core.gateway import Gateway, ReconnectWebsocket

logger = logging.getLogger("curious.voice")

class VGatewayOp(enum.IntEnum):
    IDENTIFY = 0
    SELECT_PROTOCOL = 1
    READY = 2
    HEARTBEAT = 3
    SESSION_DESCRIPTION = 4
    SPEAKING = 5
    HEARTBEAT_ACK = 6
    HELLO = 8


@async_thread(daemon=True)
def _heartbeat_loop(gw: 'VoiceGateway', heartbeat_interval: float):
    """
    Heartbeat looper that loops and sends heartbeats to the gateway.

    :param gw: The gateway to handle.
    """
    # async threads!
    logger.debug("Sending initial heartbeat.")
    AWAIT(gw.send_heartbeat())
    while True:
        # this is similar to the normal threaded event waiter
        # it will time out after heartbeat_interval seconds.
        try:
            AWAIT(curio.timeout_after(heartbeat_interval, gw._stop_heartbeating.wait()))
        except curio.TaskTimeout:
            pass
        else:
            break

        try:
            AWAIT(gw.send_heartbeat())
        except ReconnectWebsocket:
            break


class VoiceGateway(object):
    """
    A special type of websocket gateway that operates on the Voice connection.
    """

    GATEWAY_VERSION = 3

    def __init__(self, session_id: str, token: str, endpoint: str, user_id: str, guild_id: str):
        #: The current websocket object.
        self.websocket = None  # type: Websocket

        self._open = False

        self.session_id = session_id
        self.token = token
        self.user_id = user_id
        self.guild_id = guild_id
        self.sequence = 0
        self.ssrc_mapping = {}

        #: The main gateway object.
        self.main_gateway = None  # type: Gateway

        #: The current threading event used to signal if we need to stop heartbeating.
        self._stop_heartbeating = curio.Event()

        #: Voice server stuff
        self.endpoint = endpoint
        self.port = None  # type: int
        self.ssrc = None  # type: int
        self.secret_key = b""

        self._ready = curio.Event()
        self._got_secret_key = curio.Event()

    async def connect(self, url: str):
        """
        Connects the client to the websocket.

        This does not do ANY handshake steps.
        :param url: The URL to connect to.
        """
        logger.info("Opening connection to {}".format(url))
        self.websocket = await connect_websocket(url)
        logger.info("Connected to gateway!")
        self._open = True
        return self.websocket

    async def _send(self, data):
        """
        Sends some data.

        This will actually send the data down the websocket, unlike `send` which only pretends to.
        """
        try:
            await self.websocket.send_message(data)
        except WebsocketClosed:
            await self._close()

    def get_heartbeat(self):
        """
        :return: A heartbeat packet.
        """
        return {
            "op": int(VGatewayOp.HEARTBEAT),
            # ms time for v2/v3
            "d": int(round(time.time() * 1000))
        }

    async def send_heartbeat(self):
        """
        Sends a heartbeat.
        """
        hb = self.get_heartbeat()
        return await self._send_json(hb)

    def _send_json(self, payload: dict):
        """
        Sends a JSON payload down the websocket.

        This is a private method - use :meth:`send` instead.

        :param payload: The payload to send.
        """
        data = json.dumps(payload)
        return self.send(data)

    async def send(self, data: typing.Any):
        """
        Enqueues some data to be sent down the websocket.
        """
        await self.websocket.send_message(data)

    # Sending methods
    async def send_identify(self):
        """
        Sends an IDENTIFY payload.
        """
        payload = {
            "op": VGatewayOp.IDENTIFY,
            "d": {
                "server_id": str(self.guild_id),
                "user_id": str(self.user_id),
                "session_id": str(self.session_id),
                "token": str(self.token)
            }
        }
        await self._send_json(payload)

    async def send_select_protocol(self, ip: str, port: int):
        """
        Sends a SELECT PROTOCOL packet.
        """
        payload = {
            "op": VGatewayOp.SELECT_PROTOCOL,
            "d": {
                "protocol": "udp",
                "data": {
                    "address": ip,
                    "port": port,
                    "mode": "xsalsa20_poly1305"
                }
            }
        }

        await self._send_json(payload)

    async def send_speaking(self, speaking: bool = True):
        """
        Sends a SPEAKING packet.
        """
        payload = {
            "op": VGatewayOp.SPEAKING,
            "d": {
                "speaking": speaking,
                "delay": 0
            }
        }

        await self._send_json(payload)

    async def _start_heartbeating(self, heartbeat_interval: float) -> threading.Thread:
        """
        Starts the heartbeat thread.
        :param heartbeat_interval: The number of seconds between each heartbeat.
        """
        if not self._stop_heartbeating.is_set():
            # cancel some poor thread elsewhere
            await self._stop_heartbeating.set()
            del self._stop_heartbeating

        self._stop_heartbeating = curio.Event()

        # dont reference the task - it'll die by itself
        task = await curio.spawn(_heartbeat_loop(self, heartbeat_interval), daemon=True)

        return task

    async def _close(self):
        if not self.websocket.closed:
            await self.websocket.close(code=1000, reason="Client disconnected")

        self._open = False
        await self._stop_heartbeating.set()

    @classmethod
    async def from_gateway(cls, gw: Gateway, guild_id: int, channel_id: int) -> 'VoiceGateway':
        """
        Opens a new voice gateway connection from an existing Gateway connection.

        :param gw: The existing :class:`.Gateway` to create from.
        :param guild_id: The guild ID we're opening to.
        :param channel_id: The channel ID we're connecting to.
        """
        # Send our VOICE_STATE_UPDATE to tell Discord to start connecting us to voice.
        await gw.send_voice_state_update(guild_id, channel_id)
        state = await gw.state.wait_for_voice_data(guild_id)
        user_id = gw.state._user.id

        # Look up the gateway's IP address.
        endpoint = state["endpoint"]
        endpoint, port = endpoint.split(":")
        endpoint = await curio.abide(socket.gethostbyname, endpoint)

        obb = cls(token=state["token"], session_id=state["session_id"],
                  endpoint=endpoint,
                  user_id=user_id, guild_id=guild_id)
        # Open our connection to the voice websocket.
        if port == "80":
            await obb.connect(f"ws://{endpoint}/?v={cls.GATEWAY_VERSION}")
        elif port == "443":
            await obb.connect(f"wss://{endpoint}/?v={cls.GATEWAY_VERSION}")
        # Send our IDENTIFY.
        logger.info("Sending IDENTIFY...")
        await obb.send_identify()

        return obb

    async def next_event(self) -> None:
        """
        Gets the next event, in decoded form.
        """
        if not self._open:
            raise WebsocketClosed(1006, reason="Connection lost")

        try:
            event = await self.websocket.next_message()
        except WebsocketClosed:
            # Close ourselves.
            await self._close()
            raise

        if isinstance(event, WebsocketBytesMessage):
            # decompress the message
            data = zlib.decompress(event.data, 15, 10490000)
            data = data.decode("utf-8")
        else:
            data = event.data

        if data is None:
            return

        event = json.loads(data)

        op = event.get("op")
        data = event.get("d")
        seq = event.get("seq")

        # Heartbeat logic, same as normal websocket
        if seq is not None:
            self.sequence = int(seq)

        # Switch based on operator.
        if op == VGatewayOp.HELLO:
            # Ignore these, they're useless for now.
            return

        elif op == VGatewayOp.READY:
            # Start heartbeating.
            heartbeat_interval = data.get("heartbeat_interval", 45000) / 1000.0
            logger.debug("Heartbeating every {} seconds.".format(heartbeat_interval))
            await self._start_heartbeating(heartbeat_interval)
            # Set our `ssrc`, `port` and `modes`.
            self.ssrc = data.get("ssrc")
            self.port = data.get("port")

            await self._ready.set()

        elif op == VGatewayOp.SESSION_DESCRIPTION:
            # Extract the secret key.
            self.secret_key = data["secret_key"]
            await self.send_speaking()
            await self._got_secret_key.set()

        elif op == VGatewayOp.SPEAKING:
            # build a cache of user_id -> ssrc
            user_id = int(data.get("user_id"))
            ssrc = data.get("ssrc")
            self.ssrc_mapping[ssrc] = user_id

        elif op == VGatewayOp.HEARTBEAT:
            # silence
            pass

        elif op == VGatewayOp.HEARTBEAT_ACK:
            # suppress
            pass

        else:
            logger.warning("Unhandled event: {}".format(op))



"""
Decorators to annotate function objects.
"""
from typing import List

from curious.commands.utils import get_description


def command(*,
            name: str = None, description: str = None,
            aliases: List[str] = None, **kwargs):
    """
    Marks a function as a command. This annotates the command with some attributes that allow it
    to be invoked as a command.

    This decorator can be invoked like this:
    .. code-block:: python3

        @command()
        async def ping(self, ctx):
            await ctx.channel.send("Ping!")

    :param name: The name of the command. If this is not specified, it will use the name of the \
        function object.
    :param description: The description of the command. If this is not specified, it will use the \
        first line of the docstring.
    :param aliases: A list of aliases for this command.
    :param kwargs: Anything to annotate the command with.
    """

    # wrapper function that actually marks the object
    def inner(func):
        func.is_cmd = True
        func.cmd_name = name or func.__name__
        func.cmd_description = description or get_description(func)
        func.cmd_aliases = aliases or []
        func.cmd_subcommand = False
        func.cmd_subcommands = []
        func.cmd_parent = None
        func.cmd_conditions = getattr(func, "cmd_conditions", [])

        # annotate command object with any extra
        for ann_name, annotation in kwargs.items():
            ann_name = "cmd_" + name
            setattr(func, ann_name, annotation)

        func.subcommand = _subcommand(func)
        return func

    return inner


def condition(cbl):
    """
    Adds a condition to a command.

    This will add the callable to ``cmd_conditions`` on the function.
    """
    def inner(func):
        if not hasattr(func, "cmd_conditions"):
            func.cmd_conditions = []

        func.cmd_conditions.append(cbl)
        return func

    return inner


def _subcommand(parent):
    """
    Decorator factory set on a command to produce subcommands.
    """
    def inner(**kwargs):
        # MULTIPLE LAYERS
        def inner_2(func):
            cmd = command(**kwargs)(func)
            cmd.cmd_subcommand = True
            cmd.cmd_parent = parent
            parent.cmd_subcommands.append(cmd)
            return cmd

        return inner_2

    return inner




"""
Defines commands-specific exceptions.
"""


class CommandsError(Exception):
    pass


class ConditionsFailedError(CommandsError):
    """
    Raised when conditions fail for a command.
    """
    def __init__(self, ctx, check):
        self.ctx = ctx
        self.conditions = check

    def __repr__(self):
        if isinstance(self.conditions, list):
            return f"The conditions for `{self.ctx.command_name}` failed."

        return f"The condition `{self.conditions.__name__}` for `{self.ctx.command_name}` failed." \

    __str__ = __repr__


class MissingArgumentError(CommandsError):
    """
    Raised when a command is missing an argument.
    """
    def __init__(self, ctx, arg):
        self.ctx = ctx
        self.arg = arg

    def __repr__(self):
        return f"Missing required argument `{self.arg}` in `{self.ctx.command_name}`."

    __str__ = __repr__


class CommandInvokeError(Exception):
    """
    Raised when a command has an error during invokation.
    """
    def __init__(self, ctx):
        self.ctx = ctx

    def __repr__(self):
        return f"Command {self.ctx.command_name} failed to invoke with error `{self.__cause__}`."

    __str__ = __repr__


class ConversionFailedError(Exception):
    """
    Raised when conversion fails.
    """
    def __init__(self, ctx, arg: str, to_type: type):
        self.ctx = ctx
        self.arg = arg
        self.to_type = to_type

    def __repr__(self):
        return f"Cannot convert `{self.arg}` to type `{self.to_type.__name__}`."

    __str__ = __repr__



"""
A lower-level State that doesn't do any special object handling.
"""
import inspect

import typing

from curious.core.gateway import Gateway
from curious.core.state import State


class PureDispatchState(State):
    """
    A lower-level State that doesn't do any special object handling.
    This state allows you to pass JSON data straight to the event handlers registered on the Client instance.

    To use this instead of the base high-level state, you have to pass this as a class to the Client instance:

    .. code:: python

        my_client = Client(state_klass=PureDispatchState)
    """

    def _fake_handle(self, event_name: str):
        """
        Returns a function that can pretend to handle the event, when all it does is dispatch the raw data.

        :param event_name: The event name we're handling.
        """

        async def _inner(gw: Gateway, event_data: dict):
            await self.client.dispatch(event_name, event_data, gateway=gw)

        return _inner

    def __getattribute__(self, item):
        # Intercept any `handle_` things.
        if not item.startswith("handle_"):
            return super().__getattribute__(item)

        return self._fake_handle(item.split("handle_")[1])


class CallbackState(PureDispatchState):
    """
    An even lower-level State that invokes a single callback when an event is received.

    This callback must have the signature of (gw: Gateway, event: str, data: dict) -> None.

    This state can be passed directly into a Gateway instance to be used as the state instance.
    """
    def __init__(self, callback: typing.Callable[[Gateway, str, dict], None]):
        super().__init__(None)

        self.callback = callback

    def _fake_handle(self, event_name: str):
        async def _inner(gw: Gateway, event_data: dict):
            result = self.callback(gw, event_name, event_data)
            if inspect.isawaitable(result):
                result = await result

        return _inner



"""
Wrappers for Member objects (Users with guilds).

.. currentmodule:: curious.dataclasses.member
"""

import typing

from curious.dataclasses import guild as dt_guild, role as dt_role, user as dt_user, \
    voice_state as dt_vs
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.permissions import Permissions
from curious.dataclasses.presence import Game, Presence, Status
from curious.util import to_datetime


class Member(Dataclass):
    """
    A member represents somebody who is inside a guild.
    """

    __slots__ = ("_user_data", "role_ids", "joined_at", "nickname", "guild_id", "presence")

    def __init__(self, client, **kwargs):
        super().__init__(kwargs["user"]["id"], client)

        # copy user data for when the user is decached
        self._user_data = kwargs["user"]
        self._bot.state.make_user(self._user_data)

        #: An iterable of role IDs this member has.
        self.role_ids = [int(rid) for rid in kwargs.get("roles", [])]

        #: The date the user joined the guild.
        self.joined_at = to_datetime(kwargs.get("joined_at", None))

        #: The member's current nickname.
        self.nickname = kwargs.get("nick", None)

        #: The ID of the guild that this member is in.
        self.guild_id = None  # type: int

        #: The current :class:`~.Presence` of this member.
        self.presence = Presence(status=kwargs.get("status", Status.OFFLINE),
                                 game=kwargs.get("game", None))

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`~.Guild` associated with this member. 
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def voice(self) -> 'dt_vs.VoiceState':
        """
        :return: The :class:`~.VoiceState` associated with this member.
        """
        try:
            return self.guild._voice_states[self.id]
        except (AttributeError, KeyError):
            return None

    def __hash__(self) -> int:
        return hash(self.guild_id) + hash(self.user.id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Member):
            return NotImplemented

        return other.guild == self.guild and other.user == self.user

    def _copy(self):
        """
        Copies a member object.
        """
        new_object = object.__new__(self.__class__)  # type: Member
        new_object._bot = self._bot

        new_object.id = self.id
        new_object.role_ids = self.role_ids.copy()
        new_object.joined_at = self.joined_at
        new_object.guild_id = self.guild_id
        new_object.presence = self.presence
        new_object.nickname = self.nickname

        return new_object

    def __del__(self):
        try:
            self._bot.state._check_decache_user(self.id)
        except AttributeError:
            # during shutdown
            pass

    @property
    def user(self) -> 'dt_user.User':
        """
        :return: The underlying :class:`.User` for this member.
        """
        try:
            return self._bot.state._users[self.id]
        except KeyError:
            # don't go through make_user as it'll cache it
            return dt_user.User(self._bot, **self._user_data)

    @property
    def name(self) -> str:
        """
        :return: The computed display name of this user.
        """
        return self.nickname if self.nickname else self.user.username

    @property
    def mention(self) -> str:
        """
        :return: A string that mentions this member. 
        """
        if self.nickname:
            return "<@!{}>".format(self.id)

        return self.user.mention

    @property
    def status(self) -> Status:
        """
        :return: The current :class:`~.Status` of this member.
        """
        return self.presence.status if self.presence else Status.OFFLINE

    @property
    def game(self) -> Game:
        """
        :return: The current :class:`~.Game` this member is playing.
        """
        if not self.presence:
            return None

        if self.presence.status == Status.OFFLINE:
            return None

        return self.presence.game

    @property
    def roles(self) -> 'typing.Iterable[dt_role.Role]':
        """
        :return: A list of :class:`~.Role` that this member has. 
        """
        if not self.guild:
            return None

        roles = []
        for id in self.role_ids:
            try:
                roles.append(self.guild.roles[id])
            except (KeyError, AttributeError):
                pass

        return sorted(roles)

    @property
    def colour(self) -> int:
        """
        :return: The computed colour of this user.
        """
        roles = sorted(self.roles, reverse=True)
        # NB: you can abuse discord and edit the defualt role's colour
        # so explicitly check that it isn't the default role, and make sure it has a colour
        # in order to get the correct calculated colour
        roles = filter(lambda role: not role.is_default_role and role.colour, roles)
        try:
            return next(roles).colour
        except StopIteration:
            return 0

    @property
    def top_role(self) -> 'dt_role.Role':
        """
        :return: This member's top-most :class:`~.Role`.
        """
        try:
            return next(iter(sorted(self.roles, reverse=True)), self.guild.default_role)
        except AttributeError:
            return None

    @property
    def guild_permissions(self) -> Permissions:
        """
        :return: The calculated guild permissions for a member.
        """
        if self == self.guild.owner:
            return Permissions.all()

        bitfield = 0
        # add the default roles
        bitfield |= self.guild.default_role.permissions.bitfield
        for role in self.roles:
            bitfield |= role.permissions.bitfield

        permissions = Permissions(bitfield)
        if permissions.administrator:
            return Permissions.all()

        return permissions

    # Member methods.
    async def send(self, content: str, *args, **kwargs):
        """
        Sends a message to a member in DM.

        This is a shortcut for :meth:`.User.send`.
        """
        return await self.user.send(content, *args, **kwargs)

    async def ban(self, delete_message_days: int = 7):
        """
        Bans this member from the guild.

        :param delete_message_days: The number of days of messages to delete.
        """
        return await self.guild.ban(self, delete_message_days=delete_message_days)

    async def kick(self):
        """
        Kicks this member from the guild.
        """
        return await self.guild.kick(self)

    async def add_roles(self, *roles: 'typing.Iterable[dt_role.Role]'):
        """
        Adds roles to this member.

        For more documentation, see :meth:`~.Guild.add_roles`.

        :param roles: The list of roles to add.
        """
        return await self.guild.add_roles(self, *roles)

    async def remove_roles(self, *roles: 'typing.Iterable[dt_role.Role]'):
        """
        Removes roles from this member.

        For more documentation, see :meth:`~.Guild.remove_roles`.

        :param roles: The list of roles to remove.
        """
        return await self.guild.remove_roles(self, *roles)

    async def change_nickname(self, new_nickname: typing.Union[str, None]):
        """
        Changes the nickname of this member.

        :param new_nickname: The nickname to change to, None to remove the nickname.
        """
        return await self.guild.change_nickname(self, new_nickname)



"""
Converter methods.
"""
from curious.commands.exc import ConversionFailedError
from curious.dataclasses.channel import Channel
from curious.dataclasses.member import Member


def convert_member(ctx, arg: str) -> Member:
    """
    Converts an argument into a Member.
    """
    if arg.startswith("<@") and arg.endswith(">"):
        # Parse the mention out
        id = arg[2:-1]
        if id[0] == "!":  # nicknames
            id = id[1:]

        try:
            id = int(id)
        except ValueError:
            raise ConversionFailedError(ctx, arg, Member)

        member = ctx.guild.members.get(id)
        if not member:
            raise ConversionFailedError(ctx, arg, Member)
    else:
        member = ctx.guild.find_member(arg)
        if not member:
            raise ConversionFailedError(ctx, arg, Member)

    return member


def convert_channel(ctx, arg: str) -> Channel:
    """
    Converts an argument into a Channel.
    """
    if arg.startswith("<#") and arg.endswith(">"):
        id = arg[2:-1]

        try:
            id = int(id)
        except ValueError:
            raise ConversionFailedError(ctx, arg, Channel)

        channel = ctx.guild.channels.get(id)
        if not channel:
            raise ConversionFailedError(ctx, arg, Channel)
    else:
        try:
            channel = next(filter(lambda c: c.name == arg, ctx.guild.channels.values()), None)
            if channel is None:
                channel = ctx.guild.channels.get(int(arg))
        except (StopIteration, ValueError):
            raise ConversionFailedError(ctx, arg, Channel)
        else:
            if channel is None:
                raise ConversionFailedError(ctx, arg, Channel)

    return channel


def convert_int(ctx, arg: str) -> int:
    """
    Converts an argument into an integer.
    """
    try:
        return int(arg)
    except ValueError as e:
        raise ConversionFailedError(ctx, arg, int) from e


def convert_float(ctx, arg: str) -> float:
    """
    Converts an argument into a float.
    """
    try:
        return float(arg)
    except ValueError as e:
        raise ConversionFailedError(ctx, arg, float) from e



"""
Classes that wrap objects returned by Discord.

A **dataclass** is a class that contains data and allows the user to edit classes. These classes are stateful, 
as in they store state which allows the user to edit them as appropriate.

.. currentmodule:: curious.dataclasses

.. autosummary::
    :toctree: dataclasses

    appinfo
    bases
    channel
    embed
    emoji
    guild
    invite
    member
    message
    permissions
    presence
    reaction
    role
    search
    user
    voice_state
    webhook
    widget

"""



"""
Classes for plugin objects.
"""
import inspect

from curious.core import client as md_client


class Plugin(object):
    """
    Represents a plugin (a collection of events and commands under one class).
    """
    def __init__(self, client: 'md_client.Client'):
        #: The client for this plugin.
        self.client = client

    async def load(self) -> None:
        """
        Called when this plugin is loaded.

        By default, this does nothing. It is meant to be overridden to customize behaviour.
        """
        pass

    async def unload(self) -> None:
        """
        Called when this plugin is unloaded.

        By default, this does nothing. It is meant to be overridden to customize behaviour.
        """

    def _get_commands(self) -> list:
        """
        Gets the commands for this plugin.
        """
        return [i[1] for i in inspect.getmembers(self, predicate=lambda i: hasattr(i, "is_cmd"))]



"""
The main client class.

This contains a definition for :class:`.Client` which is used to interface primarily with Discord.

.. currentmodule:: curious.core.client
"""

import collections
import enum
import inspect
import logging
import traceback
import typing
from types import MappingProxyType

import curio
from asyncwebsockets import WebsocketClosed
from curio.monitor import MONITOR_HOST, MONITOR_PORT, Monitor
from curio.task import TaskGroup

from curious.core.event import EventManager, event as ev_dec
from curious.core.httpclient import HTTPClient
from curious.dataclasses import channel as dt_channel, guild as dt_guild, member as dt_member
from curious.dataclasses.appinfo import AppInfo
from curious.dataclasses.invite import Invite
from curious.dataclasses.presence import Game, Status
from curious.dataclasses.user import BotUser, User
from curious.dataclasses.webhook import Webhook
from curious.dataclasses.widget import Widget
from curious.exc import CuriousError
from curious.util import base64ify

#: A sentinel value to indicate that the client should automatically shard.
AUTOSHARD = object()

logger = logging.getLogger("curious.client")


class BotType(enum.IntEnum):
    """
    An enum that signifies what type of bot this bot is.
    
    This will tell the commands handling how to respond, as well as how to log in.
    """
    #: Regular bot. This signifies that the client should log in as a bot account.
    BOT = 1

    #: User bot. This signifies that the client should log in as a user account.
    USERBOT = 2

    # 4 is reserved

    #: No bot responses. This signifies that the client should respond to ONLY USER messages.
    ONLY_USER = 8

    #: No DMs. This signifies the bot only works in guilds.
    NO_DMS = 16

    #: No guilds. This signifies the bot only works in DMs.
    NO_GUILDS = 32

    #: Self bot. This signifies the bot only responds to itself.
    SELF_BOT = 64


class Client(object):
    """
    The main client class. This is used to interact with Discord.

    When creating a client object, you can either pass a token explicitly, or pass in in the 
    :meth:`start` call or similar.

    .. code:: python

        bot = Client("'a'")  # pass explicitly
        bot.run("'b'")  # or pass to the run call.

    """
    #: A list of events to ignore the READY status.
    IGNORE_READY = [
        "connect",
        "guild_streamed",
        "guild_chunk",
        "guild_available",
        "guild_sync"
    ]

    def __init__(self, token: str = None, *,
                 state_klass: type = None,
                 bot_type: int = (BotType.BOT | BotType.ONLY_USER)):
        """
        :param token: The current token for this bot.  
            This can be passed as None and can be initialized later.
            
        :param state_klass: The class to construct the connection state from.
        
        :param bot_type: A union of :class:`~.BotType` that defines the type of this bot.
        """
        #: The mapping of `shard_id -> gateway` objects.
        self._gateways = {}

        #: The number of shards this client has.
        self.shard_count = 0

        #: The token for the bot.
        self._token = token

        if state_klass is None:
            from curious.core.state import State
            state_klass = State

        #: The current connection state for the bot.
        self.state = state_klass(self)

        #: The bot type for this bot.
        self.bot_type = bot_type

        if self.bot_type & BotType.BOT and self.bot_type & BotType.USERBOT:
            raise ValueError("Bot cannot be a bot and a userbot at the same time")

        #: The current :class:`.EventManager` for this bot.
        self.events = EventManager()

        #: The :class:`~.HTTPClient` used for this bot.
        self.http = None  # type: HTTPClient

        #: The cached gateway URL.
        self._gw_url = None  # type: str

        #: The application info for this bot.
        #: Instance of :class:`~.AppInfo`.
        #: This will be None for user bots.
        self.application_info = None  # type: AppInfo

        self.scan_events()

    def create_http(self):
        """
        Creates the :class:`~.HTTPClient` for this bot.
        
        This requires that the token is set on ``self.token``.
        """
        if not self.http:
            self.http = HTTPClient(self._token, bot=bool(self.bot_type & BotType.BOT))

    @property
    def user(self) -> BotUser:
        """
        :return: The :class:`~.User` that this client is logged in as.
        """
        return self.state._user

    @property
    def guilds(self) -> 'typing.Mapping[int, dt_guild.Guild]':
        """
        :return: A mapping of int -> :class:`~.Guild` that this client can see.
        """
        return self.state.guilds

    @property
    def invite_url(self):
        """
        :return: The invite URL for this bot.
        """
        return "https://discordapp.com/oauth2/authorize?client_id={}&scope=bot".format(
            self.application_info.client_id)

    @property
    def events_handled(self) -> collections.Counter:
        """
        A :class:`collections.Counter` of all events that have been handled since the bot's bootup.
        This can be used to track statistics for events.
         
        .. code-block:: python3
        
            @command()
            async def events(self, ctx: Context):
                '''
                Shows the most common events.
                '''
                
                ev = ctx.bot.events_handled.most_common(3)
                await ctx.channel.send(", ".join("{}: {}".format(*x) for x in ev)
        
        """

        c = collections.Counter()
        for gw in self._gateways.values():
            c.update(gw._dispatches_handled)

        return c

    @property
    def gateways(self):
        """
        :return: A read-only view of the current gateways for this client. 
        """
        return MappingProxyType(self._gateways)

    def find_channel(self, channel_id: int):
        """
        Finds a channel by channel ID.
        """
        return self.state.find_channel(channel_id)

    async def get_gateway_url(self) -> str:
        """
        :return: The gateway URL for this bot.
        """
        if self._gw_url:
            return self._gw_url

        self._gw_url = await self.http.get_gateway_url()
        return self._gw_url

    async def get_shard_count(self) -> int:
        """
        :return: The shard count recommended for this bot.
        """
        gw, shards = await self.http.get_shard_count()
        self._gw_url = gw

        return shards

    def guilds_for(self, shard_id: int) -> 'typing.Iterable[dt_guild.Guild]':
        """
        Gets the guilds for this shard.

        :param shard_id: The shard ID to get guilds from.
        :return: A list of :class:`Guild` that client can see on the specified shard.
        """
        return self.state.guilds_for_shard(shard_id)

    def event(self, name: str):
        """
        A convenience decorator to mark a function as an event.

        This will copy it to the events dictionary, where it will be used as an event later on.

        .. code-block:: python3
        
            @bot.event("message_create")
            async def something(ctx, message: Message):
                pass

        :param name: The name of the event.
        """

        def _inner(func):
            f = ev_dec(name)(func)
            self.events.add_event(func=f)

        return _inner

    def scan_events(self):
        """
        Scans this class for functions marked with an event decorator.
        """

        def _pred(f):
            if not hasattr(f, "events"):
                return False

            if getattr(f, "scan", False):
                return True

            return False

        for _, item in inspect.getmembers(self, predicate=_pred):
            logger.info("Registering event function {} for events {}".format(_, item.events))
            self.events.add_event(item)

    # rip in peace old fire_event
    # 2016-2017
    # broke my pycharm
    async def fire_event(self, event_name: str, *args, **kwargs):
        """
        Fires an event.

        This actually passes the arguments to :meth:`.EventManager.fire_event`.
        """
        gateway = kwargs.get("gateway")
        if not self.state.is_ready(gateway.shard_id):
            if event_name not in self.IGNORE_READY and not event_name.startswith("gateway_"):
                return

        return await self.events.fire_event(event_name, *args, **kwargs, client=self)

    async def wait_for(self, *args, **kwargs):
        """
        Shortcut for :meth:`.EventManager.wait_for`.
        """
        return await self.events.wait_for(*args, **kwargs)

    # Gateway functions
    async def change_status(self, game: Game = None, status: Status = Status.ONLINE,
                            afk: bool = False,
                            shard_id: int = 0, *, sync: bool = False):
        """
        Changes the bot's current status.

        :param game: The game object to use. None for no game.
        :param status: The new status. Must be a :class:`~.Status` object.
        :param afk: Is the bot AFK? Only useful for userbots.
        :param shard_id: The shard to change your status on.
        :param sync: Sync status with other clients? Only useful for userbots.
        """
        if not self.user.bot and sync:
            # update `status` key of settings
            await self.user.settings.update(status=status.value)

        gateway = self._gateways[shard_id]
        return await gateway.send_status(game, status, afk=afk)

    # HTTP Functions
    async def edit_profile(self, *,
                           username: str = None,
                           avatar: bytes = None,
                           password: str = None):
        """
        Edits the profile of this bot.

        The user is **not** edited in-place - instead, you must wait for the `USER_UPDATE` event to
        be fired on the websocket.

        :param username: The new username of the bot.
        :param avatar: The bytes-like object that represents the new avatar you wish to use.
        :param password: The password to use. Only for user accounts.
        """
        if not self.user.bot and password is None:
            raise ValueError("Password must be passed for user bots")

        if username:
            if any(x in username for x in ('@', ':', '```')):
                raise ValueError("Username must not contain banned characters")

            if any(x == username for x in ("discordtag", "everyone", "here")):
                raise ValueError("Username cannot be a banned username")

            if not 2 <= len(username) <= 32:
                raise ValueError("Username must be 2-32 characters")

        if avatar:
            avatar = base64ify(avatar)

        await self.http.edit_user(username, avatar, password)

    async def edit_avatar(self, path: str):
        """
        A higher-level way to change your avatar.
        This allows you to provide a path to the avatar file instead of having to read it in 
        manually.

        :param path: The path-like object to the avatar file.
        """
        with open(path, 'rb') as f:
            return await self.edit_profile(avatar=f.read())

    async def get_user(self, user_id: int) -> User:
        """
        Gets a user by ID.

        :param user_id: The ID of the user to get.
        :return: A new :class:`~.User` object.
        """
        try:
            return self.state._users[user_id]
        except KeyError:
            u = self.state.make_user(await self.http.get_user(user_id))
            # decache it if we need to
            self.state._check_decache_user(u.id)
            return u

    async def get_application(self, application_id: int) -> AppInfo:
        """
        Gets an application by ID.

        :param application_id: The client ID of the application to fetch.
        :return: A new :class:`~.AppInfo` object corresponding to the application.
        """
        data = await self.http.get_app_info(application_id=application_id)
        appinfo = AppInfo(self, **data)

        return appinfo

    async def get_webhook(self, webhook_id: int) -> Webhook:
        """
        Gets a webhook by ID.

        :param webhook_id: The ID of the webhook to get.
        :return: A new :class:`~.Webhook` object.
        """
        return self.state.make_webhook(await self.http.get_webhook(webhook_id))

    async def get_invite(self, invite_code: str, *,
                         with_counts: bool = True) -> Invite:
        """
        Gets an invite by code.

        :param invite_code: The invite code to get.
        :param with_counts: Return the approximate counts for this invite?
        :return: A new :class:`~.Invite` object.
        """
        return Invite(self, **(await self.http.get_invite(invite_code, with_counts=with_counts)))

    async def get_widget(self, guild_id: int) -> Widget:
        """
        Gets a widget from a guild.
        
        :param guild_id: The ID of the guild to get the widget of. 
        :return: A :class:`~.Widget` object.
        """
        data = await self.http.get_widget_data(guild_id)
        return Widget(self, **data)

    # download_ methods
    async def download_guild_member(self, guild_id: int, member_id: int) -> 'dt_member.Member':
        """
        Downloads a :class:`~.Member` over HTTP.
        
        .. warning::
            
            The :attr:`~.Member.roles` and similar fields will be empty when downloading a Member, 
            unless the guild was in cache.
        
        :param guild_id: The ID of the guild which the member is in. 
        :param member_id: The ID of the member to get.
        :return: The :class:`~.Member` object downloaded.
        """
        member_data = await self.http.get_guild_member(guild_id=guild_id, member_id=member_id)
        member = dt_member.Member(self, **member_data)
        # this is enough to pick up the cache
        member.guild_id = guild_id

        # manual refcounts :ancap:
        self.state._check_decache_user(member.id)

        return member

    async def download_guild_members(self, guild_id: int, *,
                                     after: int = None, limit: int = 1000,
                                     get_all: bool = True) -> 'typing.Iterable[dt_member.Member]':
        """
        Downloads the members for a :class:`~.Guild` over HTTP.
        
        .. warning::
        
            This can take a long time on big guilds.
        
        :param guild_id: The ID of the guild to download members for.
        :param after: The member ID after which to get members for.
        :param limit: The maximum number of members to return. By default, this is 1000 members.
        :param get_all: Should *all* members be fetched?
        :return: An iterable of :class:`~.Member`.
        """
        member_data = []
        if get_all is True:
            last_id = 0
            while True:
                next_data = await self.http.get_guild_members(guild_id=guild_id, limit=limit,
                                                              after=last_id)
                # no more members to get
                if not next_data:
                    break

                member_data.extend(next_data)
                # if there's less data than limit, we are finished downloading members
                if len(next_data) < limit:
                    break

                last_id = member_data[-1]["user"]["id"]
        else:
            next_data = await self.http.get_guild_members(guild_id=guild_id, limit=limit,
                                                          after=after)
            member_data.extend(next_data)

        # create the member objects
        members = []
        for datum in member_data:
            m = dt_member.Member(self, **datum)
            m.guild_id = guild_id
            members.append(m)

        return members

    async def download_channels(self, guild_id: int) -> 'typing.List[dt_channel.Channel]':
        """
        Downloads all the :class:`~.Channel` for a Guild.
        
        :param guild_id: The ID of the guild to download channels for. 
        :return: An iterable of :class:`~.Channel` objects.
        """
        channel_data = await self.http.get_guild_channels(guild_id=guild_id)
        channels = []
        for datum in channel_data:
            channel = dt_channel.Channel(self, **datum)
            channel.guild_id = guild_id
            channels.append(channel)

        return channels

    async def download_guild(self, guild_id: int, *,
                             full: bool = False) -> 'dt_guild.Guild':
        """
        Downloads a :class:`~.Guild` over HTTP.
        
        .. warning::
        
            If ``full`` is True, this will fetch and fill ALL objects of the guild, including 
            channels and members. This can take a *long* time if the guild is large.
        
        :param guild_id: The ID of the Guild object to download. 
        :param full: If all extra data should be downloaded alongside it.
        :return: The :class:`~.Guild` object downloaded.
        """
        guild_data = await self.http.get_guild(guild_id)
        # create the new guild using the data specified
        guild = dt_guild.Guild(self, **guild_data)
        guild.unavailable = False

        # update the guild store
        self.state._guilds[guild_id] = guild

        if full:
            # download all of the members
            members = await self.download_guild_members(guild_id=guild_id, get_all=True)
            # update the `_members` dict
            guild._members = {m.id: m for m in members}

            # download all of the channels
            channels = await self.download_channels(guild_id=guild_id)
            guild._channels = {c.id: c for c in channels}

        return guild

    # Utility functions
    async def connect(self, token: str = None, shard_id: int = 1,
                      *, large_threshold: int = 250, **kwargs):
        """
        Connects the bot to the gateway.

        This will NOT poll for events - only open a websocket connection!
        """
        from curious.core.gateway import Gateway

        if token:
            self._token = token

        self.create_http()

        if not self.application_info and self.bot_type & BotType.BOT:
            self.application_info = AppInfo(self, **(await self.http.get_app_info(None)))

        gateway_url = await self.get_gateway_url()
        self._gateways[shard_id] = await Gateway.from_token(self._token, self.state, gateway_url,
                                                            shard_id=shard_id,
                                                            shard_count=self.shard_count,
                                                            large_threshold=large_threshold)

        return self

    async def poll(self, shard_id: int):
        """
        Polls the gateway for the next event.

        :param shard_id: The shard ID of the gateway to shard.
        """
        from curious.core.gateway import ReconnectWebsocket
        gw = self._gateways[shard_id]
        while True:
            try:
                await gw.next_event()
            except WebsocketClosed as e:
                # Try and handle the close.
                if e.reason == "Client closed connection":
                    # internal
                    return

                if e.code in [1000, 4007] or gw.session_id is None:
                    logger.info("Shard {} disconnected with code {}, "
                                "creating new session".format(shard_id, e.code))
                    self.state._reset(gw.shard_id)
                    await gw.reconnect(resume=False)
                elif e.code not in (4004, 4011):
                    # Try and RESUME.
                    logger.info("Shard {} disconnected with close code {}, reason {}, "
                                "attempting a reconnect.".format(shard_id, e.code, e.reason))
                    await gw.reconnect(resume=True)
                else:
                    raise
            except ReconnectWebsocket:
                # We've been told to reconnect, try and RESUME.
                await gw.reconnect(resume=True)

    async def boot_shard(self, shard_id: int, shard_count: int = None,
                         **kwargs) -> curio.Task:
        """
        Boots a single gateway shard.
        
        This can be used to run multiple clients instead of having a single client shard.
        
        :param shard_id: The shard ID to boot. 
        :param shard_count: The number of shards being created.
        :return: The :class:`curio.Task` that represents the polling loop.
        """
        if shard_count:
            self.shard_count = shard_count

        logger.info(f"Spawning shard {shard_id}")
        await self.connect(token=self._token, shard_id=shard_id, **kwargs)
        if "task_group" in kwargs:
            spawn = kwargs["task_group"].spawn
        else:
            spawn = curio.spawn

        t = await spawn(self.poll(shard_id))
        t.task_local_storage["shard_id"] = {"id": shard_id}
        return t

    async def start(self, shards: int = 1, **kwargs):
        """
        Starts the gateway polling loop.

        This is a convenience method that polls on all the shards at once.
        This will **only reboot safely returned shards.** Erroring shards won't be rebooted.
        """
        logger.info("Starting bot with {} shards.".format(shards))
        self.shard_count = shards
        results = []

        async with TaskGroup(name="shard waiter") as g:
            for shard_id in range(0, shards):
                shard_listener = await self.boot_shard(shard_id, task_group=g, **kwargs)
                logger.info("Sleeping for 5 seconds between shard creation.")
                if shard_id < shards - 1:
                    await curio.sleep(5)

            while True:
                task = await g.next_done()  # type: curio.Task
                if task is None:
                    break

                try:
                    result = await task.join()
                except Exception as e:
                    result = e
                    # format custom exception
                    exc = traceback.format_exception(None, e.__cause__, e.__cause__.__traceback__)
                    exc = ''.join(exc)

                    logger.error("Shard {} crashed, not rebooting it.\n{}"
                                 .format(task.task_local_storage["shard_id"]["id"], exc))
                else:
                    # reboot it
                    logger.warning("Rebooting shard {}.".format(task.id))
                    shard_id = task.task_local_storage["shard_id"]["id"]
                    t = await self.boot_shard(shard_id=shard_id)
                    t.task_local_storage["shard_id"] = {"id": shard_id}
                    # add the shard waiter task
                    g.add_task(t)
                finally:
                    results.append(result)

        # if we're still here, cancel a bunch of stuff
        for gateway in self._gateways.values():
            await gateway.close()

        return results

    async def start_autosharded(self, token: str = None, **kwargs):
        """
        Starts the bot with an automatically set number of shards.
        """
        if token:
            self._token = token

        self.create_http()

        shards = await self.get_shard_count()
        self.shard_count = shards
        await self.start(shards=shards, **kwargs)

    async def _cleanup(self):
        """
        Performs cleanup.
        """
        for gateway in self._gateways.values():
            await gateway.close()

    def run(self, token: str = None, shards: typing.Union[int, object] = 1, *,
            monitor_host: str = MONITOR_HOST, monitor_port: int = MONITOR_PORT,
            **kwargs):
        """
        Runs your bot with Curio with the monitor enabled.

        :param token: The token to run with.
        :param shards: The number of shards to run.
            If this is None, the bot will autoshard.
            
        :param monitor_host: The host of the Curio monitor to use.
        :param monitor_port: The port of the Curio monitor to use.
        """
        if token is not None:
            self._token = token

        if not self.bot_type & BotType.BOT and shards != 1:
            raise CuriousError("Cannot start user bots in sharded mode")

        kernel = curio.Kernel()
        monitor = Monitor(kernel, monitor_host, monitor_port)
        if shards == AUTOSHARD or shards is None:
            coro = self.start_autosharded(token, **kwargs)
        else:
            coro = self.start(shards=shards, **kwargs)

        try:
            return kernel.run(coro, shutdown=True)
        except (KeyboardInterrupt, EOFError):
            kernel._crashed = False
            return kernel.run(self._cleanup())

    @classmethod
    def from_token(cls, token: str = None):
        """
        Starts a bot from a token object.

        :param token: The token to use for the bot.
        """
        bot = cls(token)
        return bot.run()



"""
Contains the class for the commands manager for a client.
"""
import importlib
import inspect
import logging
import sys
import traceback
import typing
from collections import defaultdict

import curio

from curious.commands.context import Context
from curious.commands.exc import CommandsError
from curious.commands.help import help_command
from curious.commands.plugin import Plugin
from curious.commands.utils import prefix_check_factory
from curious.core import client as md_client
from curious.core.event import EventContext, event
from curious.dataclasses.message import Message

logger = logging.getLogger("curious.commands.manager")


class CommandsManager(object):
    """
    A manager that handles commands for a client.

    First, you need to create the manager and attach it to a client:
    .. code-block:: python3

        # form 1, automatically register with the client
        manager = CommandsManager.with_client(bot)

        # form 2, manually register
        manager = CommandsManager(bot)
        manager.register_events()

    This is required to add the handler events to the client.

    Next, you need to register a message check handler. This is a callable that is called for
    every message to try and extract the command from a message, if it matches.
    By default, the manager provides an easy way to use a simple command prefix:
    .. code-block:: python3

        # at creation time
        manager = CommandsManager(bot, command_prefix="!")

        # or set it on the manager
        manager.command_prefix = "!"

    At this point, the command prefix will be available on the manager with either
    :attr:`.Manager.command_prefix` or :attr:`.Manager.message_check.prefix`.

    If you need more complex message checking, you can use ``message_check``:
    .. code-block:: python3

        manager = CommandsManager(bot, message_check=my_message_checker)
        # or
        manager.message_check = my_message_checker

    Finally, you can register plugins or modules containing plugins with the manager:
    .. code-block:: python3

        @bot.event("ready")
        async def load_plugins(ctx: EventContext):
            # load plugin explicitly
            await manager.load_plugin(PluginClass, arg1)
            # load plugins from a module
            await manager.load_plugins_from("my.plugin.module")

    You can also add free-standing commands that aren't bound to a plugin with
    :meth:`.CommandsManager.add_command`:
    .. code-block:: python3

        @command
        async def ping(ctx: CommandsContext):
            await ctx.channel.send(content="Pong!")

        manager.add_command(ping)

    These will then be available to the client.
    """

    def __init__(self, client: 'md_client.Client', *,
                 message_check=None, command_prefix: str = None):
        """
        :param client: The :class:`.Client` to use with this manager.
        :param message_check: The message check function for this manager.

            This should take two arguments, the client and message, and should return either None
            or a 2-item tuple:
              - The command word matched
              - The tokens after the command word
        """
        if message_check is None and command_prefix is None:
            raise ValueError("Must provide one of message_check or command_prefix")

        #: The client for this manager.
        self.client = client

        if message_check is None:
            message_check = prefix_check_factory(command_prefix)

        #: The message check function for this manager.
        self.message_check = message_check

        #: A dictionary mapping of <plugin name> -> <plugin> object.
        self.plugins = {}

        #: A dictionary of stand-alone commands, i.e. commands not associated with a plugin.
        self.commands = {}

        self._module_plugins = defaultdict(lambda: [])

    @classmethod
    def with_client(cls, client: 'md_client.Client', **kwargs):
        """
        Creates a manager and automatically registers events.
        """
        obb = cls(client=client, **kwargs)
        obb.register_events()
        return obb

    def register_events(self) -> None:
        """
        Copies the events to the client specified on this manager.
        """
        self.client.events.add_event(self.handle_message)
        self.client.events.add_event(self.default_command_error)
        self.client.events.event_hooks.append(self.event_hook)

        from curious.commands.decorators import command
        self.commands["help"] = command(name="help")(help_command)

    async def load_plugin(self, klass: typing.Type[Plugin], *args,
                          module: str = None):
        """
        Loads a plugin.
        .. note::

            The client instance will automatically be provided to the Plugin's ``__init__``.

        :param klass: The plugin class to load.
        :param args: Any args to provide to the plugin.
        :param module: The module name provided with this plugin. Only used interally.
        """
        # get the name and create the plugin object
        plugin_name = getattr(klass, "plugin_name", klass.__name__)
        instance = klass(self.client, *args)

        # call load, of course
        await instance.load()

        self.plugins[plugin_name] = instance
        if module is not None:
            self._module_plugins[module].append(instance)

        return instance

    async def unload_plugin(self, klass: typing.Union[Plugin, str]):
        """
        Unloads a plugin.

        :param klass: The plugin class or name of plugin to unload.
        """
        p = None
        if isinstance(klass, str):
            p = self.plugins.pop(klass)

        for k, p in self.plugins.copy().items():
            if type(p) == klass:
                p = self.plugins.pop(k)
                break

        if p is not None:
            await p.unload()

        return p

    def _lookup_command(self, name: str):
        """
        Does a lookup in plugin and standalone commands.
        """
        if name in self.commands:
            return self.commands[name]

        for plugin in self.plugins.values():
            cmds = plugin._get_commands()

            try:
                return next(filter(lambda cmd: not cmd.cmd_subcommand and
                                   (cmd.cmd_name == name or name in cmd.cmd_aliases), cmds))
            except StopIteration:
                continue

    def get_command(self, command_name: str):
        """
        Gets a command from the internal command storage.
        If provided a string separated by spaces, a subcommand lookup will be attempted.

        :param command_name: The name of the command to lookup.
        """
        # do an immediate lookup for the first token
        sp = command_name.split(" ")
        command = self._lookup_command(sp[0])

        if command is None:
            return None

        for token in sp[1:]:
            try:
                filtered = filter(lambda cmd: cmd.cmd_name == token or token in cmd.cmd_aliases,
                                  command.cmd_subcommands)
                command = next(filtered)
            except StopIteration:
                return None

        return command

    def add_command(self, command):
        """
        Adds a command.

        :param command: A command function.
        """
        if not hasattr(command, "is_cmd"):
            raise ValueError("Commands must be decorated with the command decorator")

        self.commands[command.cmd_name] = command
        return command

    def remove_command(self, command):
        """
        Removes a command.

        :param command: The name of the command, or the command function.
        """
        if isinstance(command, str):
            return self.commands.pop(command)
        else:
            for k, p in self.commands.copy().items():
                if p == command:
                    return self.commands.pop(k)

    async def load_plugins_from(self, import_path: str):
        """
        Loads plugins from the specified module.

        :param import_path: The import path to import.
        """
        mod = importlib.import_module(import_path)

        # define the predicate for the body scanner
        def predicate(item):
            if not isinstance(item, type):
                return False

            # only accept plugin subclasses
            if not issubclass(item, Plugin):
                return False

            # ensure item is not actually Plugin
            if item == Plugin:
                return False

            # it is a plugin
            return True

        for plugin_name, plugin_class in inspect.getmembers(mod, predicate=predicate):
            await self.load_plugin(plugin_class, module=mod)

    async def unload_plugins_from(self, import_path: str):
        """
        Unloads plugins from the specified module.
        This will delete the module from sys.path.

        :param import_path: The import path.
        """
        for plugin in self._module_plugins[import_path]:
            await plugin.unload()
            self.plugins.pop(getattr(plugin, "plugin_name", "__name__"))

        del sys.modules[import_path]
        del self._module_plugins[import_path]

    async def event_hook(self, ctx: EventContext, *args, **kwargs):
        """
        The event hook for the commands manager.
        """
        taskgroup = curio.TaskGroup()
        for plugin in self.plugins.values():
            body = inspect.getmembers(plugin, predicate=lambda v: hasattr(v, "is_event"))
            for _, handler in body:
                if ctx.event_name not in handler.events:
                    continue

                await taskgroup.spawn(self.client.events._safety_wrapper(
                    handler, ctx, *args, **kwargs
                ))

        await taskgroup.join()

    async def handle_commands(self, ctx: EventContext, message: Message):
        """
        Handles commands for a message.
        """
        # step 1, match the messages
        matched = self.message_check(self.client, message)
        if inspect.isawaitable(matched):
            matched = await matched

        if matched is None:
            return None

        # deconstruct the tuple returned into more useful variables than a single tuple
        command_word, tokens = matched

        # step 2, create the new commands context
        ctx = Context(event_context=ctx, message=message)
        ctx.command_name = command_word
        ctx.tokens = tokens
        ctx.manager = self

        # step 3, invoke the context to try and match the command and run it
        await ctx.try_invoke()

    @event("command_error")
    async def default_command_error(self, ctx: Context, err: CommandsError):
        """
        Handles command errors by default.
        """
        if len(self.client.events.event_listeners.getall("command_error")) > 1:
            return

        fmtted = ''.join(traceback.format_exception(type(err), err, err.__traceback__))
        logger.error(f"Error in command!\n{fmtted}")

    @event("message_create")
    async def handle_message(self, ctx: EventContext, message: Message):
        """
        Registered as the event handler in a client for handling commands.
        """
        return await self.handle_commands(ctx, message)
